; An Emacs major mode for editing `Structured Log' files.
;
; A structured log file contains a sequence of fault/suggestion reports,
; with various fields for tracking the status of each report.
; Some of the help strings etc. are specifically for the Ergo proof tool,
; but it could fairly easily be adapted to other applications, perhaps
; by making it more table driven.
;
; Copyright (C) 1996  The University of Queensland
;
; Authors:
;   Mark Utting
;   Software Verification Research Centre
;   Department of Computer Science
;   Australia 4072
;   The University of Queensland
;   Email: marku@cs.uq.oz.au
;
; Installation:
;   Put this file into a directory that is in your Emacs load-path,
;   and add the following lines to your .emacs file:
;
;        (autoload 'slog-mode "SLog-mode"
;           "Major mode for structured log files")
;        (setq auto-mode-alist (cons '("\\.slog\$" . slog-mode)
;                                    auto-mode-alist))


(defconst slog-mode-version "$Revision: 2.4 $"
  "The version of the structured log major mode")

(defconst slog-record-separator-regexp
  "^--------"
  "A regular expression to match record separators")

(defvar slog-nextprev-regexp
  "^Descr: *"
  "*The slog-next/previous commands position the cursor on this line")

(defvar slog-outline-regexp
  "^Descr:"
  "*A regular expression to match interesting lines, for outline mode")

(defconst slog-ergo-version-regexp
  "^ *Ergo *-- *release *\\(\\S-+\\)"
  "A regular expression to match record separators")

(defconst slog-keyword-regexp
  "^\\([A-Z][a-z]+\\):"
  "A regular expression to match keywords.")

;; The master-list of all fields.
(defvar slog-fields
  '(("Id"      . "The 'Identity' field contains a number and a date.
The number should be unique throughout the whole file.
You should NOT normally change the identity number of an existing 
record, because it may be referred to by CVS logs and other places.")
    ("From"    . "The 'From' field records the name of the person
who initiated this request.  This is not necessarily the person
who entered the request into this file.  If possible, the name
should be a legal email address.  When the value of this field
is prompted for in the mini-buffer, completion is available
\(all existing 'From' names in the file).")
    ("Kind"    . slog-kind-alternatives)
    ("Urgency" . slog-urgency-alternatives)
    ("Release" . "The release number of Ergo that this report relates to.")
    ("Descr"   . "The 'Description' field should contain a one-line
summary of the request, immediately following the 'Descr:' keyword.

Further details of the request should be on the following lines,
indented by at least 8 spaces (1 tab)

For fault records, you should include sufficient details to enable
the fault to be reproduced.  For example, when reporting a fault
that occurs during an Ergo proof, consider including:

  * the name of the current theory
  * the name or statement of the theorem being proved
  * a copy of the current window (focus, relation, goal, hypotheses)
  * the current not_free_in constraints  (use 'show(allnfi).')
  * the command/tactic call that failed (with its arguments)
  * a copy (or description) of the messages that Ergo produced
    as the failure happened.")
    ("Action"  . "Each 'Action' line documents the date upon which
that action was begun/ended, the person who performed that action and
a rough estimate of how many person-hours the action took.
The actions are:

  * Acknowledged: (sending an acknowledgement message for reports
        received via email or on paper).
        If your acknowledgement message suggested a work-around, you
        may record that work-around on the following lines for others
        to read.

  * Analysed: (tracking down the source of the fault, or deciding
        whether it is useful and feasible to implement a suggestion).
        If you fixed the fault immediately, the results of your
        analysis will probably be included down in the 'Summary'
        field where you describe the fix.  Otherwise, you may
        want to record the results of your analysis on following lines.

  * Fixed: (fixing the cause of a fault, or implementing a suggestion).
        Details of the fix/implementation go in the Summary field.

  * Tested: (includes general regression testing, creation of new validation 
        tests, and checking that the fault has indeed been fixed
        or that the suggestion has achieved the intended purpose).
	Following lines may describe what features were/weren't tested.

The slog-begin (\\C-c\\C-b) and slog-end (\\C-c\\C-e) commands
fill in the begin/end information with default values, nicely aligned.
If the actual dates/names are different, please edit them until they
are correct.  The format for dates is: [D]DMonYYYY.
Dates with DD=00 mean 'sometime during that month'.  Eg. 00May1996.

READ THIS BEFORE YOU EDIT 'ACTION' LINES!
The programs that analyse action lines assume that:

  * actions that have not yet been started match ' by *$',
  * actions that have been started but not finished match ' -- *$',
  * completed actions match 'Hours:[ 0-9\.]*$'.

You can delete any 'Action:' lines that are irrelevant for
a particular request.  For example, an 'Acknowledge' action might
be deleted if the person entering the record is the same person
who initiated the request.  Similarly, if a particular action
\(eg. testing) is repeated several times, you may duplicate that
action line in order to record the different people, dates and times.
Alternatively, if several people contribute to the one action in the 
same time period, you may list their names within the same action line, 
separated by commas.")
    ("Summary" . slog-summary-alternatives)
    )
  "Defines the fields that have a fixed set of values.
This is an association list that maps each field to a help string
for that field, or (for fields with a fixed set of alternatives)
the name of an assocation list of the alternatives for that field.")

;; Details (and help strings) for each kind of `alternative' field.
(defvar slog-kind-alternatives
  '(("fault" .
     "A `fault' entry should describe a possible problem with the
      tool or something that is difficult to use/understand,
      or some part of its documentation that is incorrect or misleading.
      Some `faults' may be simply misunderstandings, but they should still
      be recorded, since improvements to the documentation may help to reduce
      such misunderstandings.
      The details associated with a fault entry should be specific enough for 
      the problem to be reproduced.")
    ("suggestion" . 
     "A `suggestion' entry records a suggestion for improving the
      tool or its documentation.") 
    ("observation" .
     "An `observation' entry covers any other kind of report that
      is not a fault or a suggestion."))
  "The recommended alternatives for the `Kind' field.")

(defvar slog-urgency-alternatives
  '(("low" .
     "Low urgency means that there is an easy work-around for the
      fault/suggestion that does not greatly inconvenience users.")
    ("medium" . 
     "Medium urgency means that although users can continue without 
      this fault/suggestion being addressed, the work-around is 
      inconvenient and time-consuming.")
    ("high" .
     "High urgency means that users cannot continue using the tool
      until this fault/suggestion has been addressed."))
  "The recommended alternatives for the `Urgency' field.")


(defvar slog-summary-alternatives
  '(
    ("prover" .
     "The fault/suggestion resulted in changes to the Ergo prover.
      If possible, please also indicate the source file(s) or procedures that 
      were changed.  Eg. prover/dbms/theory_data.ql.")
    ("globaltactic" . 
     "The fault/suggestion required changes to some of the
      global (theory-independent) Ergo tactics.  
      If possible, please also indicate the source file(s) or procedures that 
      were changed.  Eg. tactics/usefwd.ql.")
    ("theory" . 
     "The fault/suggestion required changes to some standard Ergo 
      theory (includes all theories that are released with Ergo, such as
      the Cogito theories).  
      If possible, please also indicate the theory(s) that were changed.
      Eg. schema.thy.")
    ("theorytactic"
     "The fault/suggestion required changes to tactics within
      some standard Ergo theory (includes all theory-specific tactics
      that are released with Ergo, such as the Cogito theories).  
      If possible, please also indicate the theory and tactic file(s) 
      that were changed.  Eg. schema/schema.ql, schema/testing.ql.")
    ("emacsinterface" . 
     "The fault/suggestion required changes to the Ergo-Emacs interface.")
    ("doc"
     "The fault/suggestion required changes *only* to 
      the documentation (on-line, tutorials or manuals).
      If possible, please also include the name of the documents or files
      that were changed.  Eg. doc/tutorial/tutorial.thy.")
    ("user"
     "The fault/suggestion was a mistake/misunderstanding by a user.
      Recording these may suggest improvements to the documentation
      to reduce similar problems in the future.")
    ("qu-prolog"
     "The fault/suggestion required changes to Qu-Prolog.")
    ("misc"
     "Any other cause of a fault that does not fit the other categories.")
    )
  "The recommended alternatives for the `Summary' field.
If multiple alternatives are relevant, add extra `Summary:' lines.
Each line should have the format `Summary: KeyWord FileNames...',
where KeyWord is one of the prover/theory/doc... alternatives.
If more details are required, add text on the following lines.
The basic idea is to give just enough information to enable an interested
reader to find the relevant CVS log entries reasonably easily.
")

(defvar slog-name-history
  nil
  "History list for SLog user names.")

(defvar slog-syntax-table (make-syntax-table)
  "Syntax table for the Slog mode")

(defvar slog-abbrev-table (make-abbrev-table)
  "Abbrev table for the Slog mode")

(defvar slog-mode-hooks nil
  "*Hooks run after starting Slog mode.")


;; Define the SLog menu (items are in reverse order) 
(defvar slog-menu (make-sparse-keymap "SLog Commands"))
(define-key slog-menu [doc]  '("Describe current field" . slog-describe))
(define-key slog-menu [prev]    '("Previous record" . slog-previous))
(define-key slog-menu [next]    '("Next record" . slog-next))
(define-key slog-menu [cycle]   '("Cycle to next alternative" . slog-cycle))
(define-key slog-menu [indent]  '("Indent region" . slog-indent-region))
(define-key slog-menu [end]     '("End current action" . slog-end-action))
(define-key slog-menu [begin]   '("Begin current action" . slog-begin-action))
(define-key slog-menu [observ]  '("Add new Observation" . slog-observation))
(define-key slog-menu [sugg]    '("Add new Suggestion" . slog-suggestion))
(define-key slog-menu [fault]   '("Add new Fault" . slog-fault))

;; The local keymap for the Slog mode.
(defvar slog-mode-map (make-keymap "SLog"))
(define-key slog-mode-map "\C-c\C-f" 'slog-fault)
(define-key slog-mode-map "\C-c\C-s" 'slog-suggestion)
(define-key slog-mode-map "\C-c\C-o" 'slog-observation)
(define-key slog-mode-map "\C-c\C-b" 'slog-begin-action)
(define-key slog-mode-map "\C-c\C-e" 'slog-end-action)
(define-key slog-mode-map "\C-c\C-c" 'slog-cycle)
(define-key slog-mode-map "\C-c\C-i" 'slog-indent-region)
(define-key slog-mode-map "\C-c\C-n" 'slog-next)
(define-key slog-mode-map "\C-c\C-p" 'slog-previous)
(define-key slog-mode-map "\C-c\C-d" 'slog-describe)
(define-key slog-mode-map "\C-m"     'newline-and-indent)
(define-key slog-mode-map "\t"       'slog-indent-line)
;; Link the SLog menu into the local keymap for this mode.
(define-key slog-mode-map [menu-bar slog] (cons "SLog" slog-menu))


(defun slog-mode ()
  "Major mode for editing Structured Log files.

Each log entry has the following form ({...} text is just commentary):
--------                                {8 minus signs}
Id:     99  02May1996                   {ID increments sequentially}
From:   email@somewhere
Kind:   fault                           {fault/suggestion/observation}
Urgency:medium                          {low/medium/high}
Release:4.1.6
Descr:  One line summary...
	{Followed by a full description of fault/suggestion,
	 including sufficient details to enable faults to be reproduced.
         Eg. current theory, theorem, current window, nfi's...}
Action: acknow   by marku               02May1996 -- 02May1996  Hours: 0.25
	{Optional details of a workaround, if any}
Action: analysed  by marku              02May1996 -- 02May1996  Hours: 1
	{Optional conclusions of initial analysis}
Action: fixed     by nickson            03May1996 -- 03May1996  Hours: 0.5
Action: tested    by marku              03May1996 --
	{Optional description of what was/wasn't tested.}
Summary: tactic global:postulate/1      {prover/globaltactic/
                                         theory/theorytactic/doc/user}
	{Summary of how the fault/suggestion was fixed/implemented.}
--------

You can delete any 'Action:' lines that are irrelevant.
Please indent all description lines by one tab, as shown above.

\\[slog-describe]  describes the intended use of the current field.

\\[slog-fault]  appends a new `fault' record. 
\\[slog-suggestion]  appends a new `suggestion' record.
\\[slog-observation]  appends a new `observation' record.
\\[slog-begin-action]  fills in start details of the current action line.
\\[slog-end-action]  fills in end details of the current action line.
\\[slog-cycle]  cycles to the next alternative for the current field.
  (relevant for the Kind,Urgency and Summary fields)
\\[slog-indent-region]  indents the current region by 1 tab

\\[slog-next]  moves to the next record in the log
\\[slog-previous]  moves to the previous record in the log
"
  (interactive)
  (setq mode-name "SLog"
        major-mode 'slog-mode)
  (use-local-map slog-mode-map)
  (set-syntax-table slog-syntax-table)
  (make-local-variable 'local-abbrev-table)
  (setq local-abbrev-table slog-abbrev-table)
  (setq outline-regexp slog-outline-regexp)
  (make-local-variable 'paragraph-start)
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate "^[ \t]*$")
  (setq paragraph-start 
	(concat slog-record-separator-regexp "\\|"
		slog-keyword-regexp "\\|"
		paragraph-separate))
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)  ;; is this a good idea?
  (setq fill-prefix "\t")  ;; automatically becomes buffer-local
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'slog-indent-line)
  (setq fill-column 75)
  (auto-fill-mode)
  ; (outline-minor-mode)
  (run-hooks 'slog-mode-hooks)
)


(defun slog-indent-region (top bottom)
  "Indents every line in the current region by exactly one additional tab.
This is useful for indenting text on the lines following a keyword line,
in the structured log format.  In particular, it is useful to call this
immediately after pasting a chunk of text into the log file with the mouse,
because the region will be exactly the text just pasted."
  (interactive "r")
  (indent-rigidly top bottom 8))


(defun slog-indent-line ()
  "Indents the current line according to the `Structured Log' style.
Lines that begin with a keyword are not indented.
Other lines are assumed to be continuation lines for the previous
keyword, so are indented by one tab, or more if the previous line
is indented further.
"
  (interactive)
  (let (keywordline start end col target-col)
    (beginning-of-line)
    ;; calculate current indentation of current line
    ;; (for keyword lines, the indentation is the whitespace 
    ;;  *after* the keyword)
    (setq keywordline (looking-at slog-keyword-regexp))
    (if keywordline (search-forward ":"))
    (setq start (point))       ;; start of indentation
    (skip-chars-forward " \t")
    (setq end        (point)   ;; end of the indentation on curr line
	  col        (current-column))
    ;; now calculate desired indentation, by looking at previous line.
    (setq target-col
	  (if keywordline
	      8
	    (save-excursion 
	      (forward-line -1)
	      (if (looking-at slog-keyword-regexp)
		  8
		(skip-chars-forward " \t")
		(max 8 (current-column))))))
    (if (= col target-col)
	(goto-char end)
      (delete-region start end)
      (goto-char start)
      (indent-to target-col))))


(defun slog-current-date ()
  "Returns the current date in the standard SLog format (DDMonYYYY)"
  (format-time-string "%e%b%Y"))


(defun slog-user-list ()
  "Returns a association list of all the `From:' fields in the file.
The CAR is the name and the CDR is nil."
  (let ((ulist nil) user)
    (save-excursion
      (goto-char (point-min))
      (while (search-forward-regexp "^From: *\\([a-zA-Z].+\\) *$" nil t nil)
	(setq user (buffer-substring (match-beginning 1) (match-end 1)))
	(if (not (assoc user ulist))
	    (setq ulist (cons (cons user nil) ulist)))))
    ulist))


(defun slog-insert-entry (kind)
  "Append a log entry to the file, of the given Kind"
  (let ((name (let ((completion-ignore-case t))
		(completing-read 
		 "Reported by: " 
		 (slog-user-list)
		 nil
		 nil
		 (user-login-name) 
		 'slog-name-history)))
	(lastid (save-excursion
		  (goto-char (point-max))
		  (if (search-backward-regexp "^Id: *\\([0-9]+\\)" nil t)
		      (string-to-number (buffer-substring (match-beginning 1)
							  (match-end 1)))
		    0)))
	(release (if (not (get-buffer "*ergo*"))
		     ""
		   (save-excursion
		     (set-buffer "*ergo*")
		     (save-excursion
		       (goto-char (point-max))
		       (if (search-backward-regexp slog-ergo-version-regexp 
						   nil t)
			   (buffer-substring (match-beginning 1)
					     (match-end 1))
			 "")))))
	dest)
    (goto-char (point-max))
    (skip-chars-backward " \t\n")
    (delete-region (point) (point-max))
    (insert "
--------
Id:     " (number-to-string (+ lastid 1)) 
  "  " (slog-current-date) "
From:   " name "
Kind:   " kind "
Urgency:low
Release:" release "
Descr:  ")
    (setq dest (point))
    (insert "
Action: acknow   by 
Action: analysed by 
Action: fixed    by 
Action: tested   by 
Summary: 
")
    (goto-char dest)
    (message "(One line summary, then details on following lines)")
))


(defun slog-fault ()
  "Append a new fault record to the buffer"
  (interactive)
  (slog-insert-entry "fault"))


(defun slog-suggestion ()
  "Append a new suggestion record to the buffer"
  (interactive)
  (slog-insert-entry "suggestion"))


(defun slog-observation ()
  "Append a new observation record to the buffer"
  (interactive)
  (slog-insert-entry "observation"))


(defun slog-next ()
  "Move to the next record"
  (interactive)
  (search-forward-regexp slog-record-separator-regexp)
  (search-forward-regexp slog-nextprev-regexp))


(defun slog-previous ()
  "Move to the previous record"
  (interactive)
  (search-backward-regexp slog-record-separator-regexp)
  (search-backward-regexp slog-nextprev-regexp))


(defun slog-begin-action (arg)
  "Enter default start details for the current action line.
If you type \\[universal-argument] before this command, it prompts for
the user name to insert, otherwise it inserts your login name.

It inserts the current date, so you should edit this if the action
was not started today.  All dates must be in [D]DMonYYYY format.
Dates with DD=00 mean 'sometime during that month'.  Eg. 00May1996."
  (interactive "P")
  (save-excursion
    (beginning-of-line)
    (if (not (looking-at "Action:"))
	(error "Not an `Action:' line"))
    (if (not (looking-at "Action:[ a-z]* by *$"))
	(error "This action already has start-time information")))
  (let ((name (if arg
		  (read-from-minibuffer 
		     "Login name: " 
		     nil
		     nil
		     nil
		     'slog-name-history)
		(user-login-name)))
	padding)
    (end-of-line)
    (just-one-space)
    (setq padding (make-string (max 1 (- 20 (length name))) ?\ ))
    (insert name padding (slog-current-date) " --")))


(defun slog-end-action ()
  "Enter default finishing details for the current action line.

It inserts the current date, so you should edit this if the action
was not finished today.  All dates must be in [D]DMonYYYY format.
Dates with DD=00 mean 'sometime during that month'.  Eg. 00May1996."
  (interactive)
  (save-excursion
    (beginning-of-line)
    (if (not (looking-at "Action:"))
	(error "Not an `Action:' line"))
    (if (not (looking-at "Action:[ a-z]* by.*-- *$"))
	(error "This action already has end-time information")))
  (end-of-line)
  (just-one-space)
  (insert (slog-current-date) "  Hours: "))


(defun slog-cycle (n)
  "Cycles to the Nth next alternative for the current field."
  (interactive "p")
  (let (field altlist curr begincurr endcurr next)
    (save-excursion ;; in case of errors, we leave the cursor unchanged.
      (end-of-line)
      (if (not (search-backward-regexp slog-keyword-regexp nil t))
	  (error "Cannot determine current field!"))
      (setq field (buffer-substring (match-beginning 1) (match-end 1)))
      ;; Find the appropriate assoc-list for this field.
      (setq altlist (assoc field slog-fields))
      (if (not (and altlist (symbolp (cdr altlist))))
	  (error "Current field does not have alternatives."))
      (setq altlist (symbol-value (cdr altlist)))
      ;; Now extract the current alternative.
      (search-forward-regexp slog-keyword-regexp nil t)
      (if (not (looking-at " *\\([-a-zA-Z]+\\)\\( *\\)"))
	  ;; There is no current alternative, so we set next 
	  ;; to be the first alternative in altlist.
	  (setq curr nil
		begincurr (point)
		endcurr (point)
		next (car (car altlist)))
	(setq curr (buffer-substring (match-beginning 1) (match-end 1))
	      begincurr (match-beginning 1)
	      endcurr (match-end 2))
	(if (not (assoc curr altlist))
	    (error "Current alternative is not legal"))
	;; Search through altlist to find the next alternative.
        (setq n (% (+ n (length altlist)) (length altlist)))
	(setq altlist (append altlist altlist))
	(while (and altlist (not (equal curr (car (car altlist)))))
	  (setq altlist (cdr altlist)))
	(setq next (car (nth n altlist)))))
    ;; The save-excursion code has set curr, begincurr, endcurr and next.
    ;; Replace curr by next.
    (delete-region begincurr endcurr)
    (goto-char begincurr)
    (insert next " ")))


(defun slog-describe ()
  "Describes the valid alternatives for the current field."
  (interactive)
  (let (field altlist altlistdoc)
    (save-excursion
      (end-of-line)
      (if (not (search-backward-regexp slog-keyword-regexp nil t))
	  (error "Cannot determine current field!"))
      (setq field (buffer-substring (match-beginning 1) (match-end 1)))
      ;; Find the appropriate assoc-list for this field.
      (setq altlist (assoc field slog-fields))
      (if (not altlist)
	  (error "Current field does not have alternatives."))
      (if (symbolp (cdr altlist))
	  (setq altlistdoc
		   (documentation-property (cdr altlist)
					   'variable-documentation)
		altlist (symbol-value (cdr altlist)))
	(setq altlistdoc (cdr altlist)
	      altlist    nil)))
    (let ((fmt (function (lambda (pair)
			   (princ (format "'%s': " (car pair)))
			   (princ (cdr pair))
			   (princ "\n\n")))))
      (with-output-to-temp-buffer "*Help*"
	(princ altlistdoc)
	(princ "\n\n")
	(mapcar fmt altlist)
	(print-help-return-message)))))
  


(provide 'slog)
