%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: comprule.ql,v 1.43 2001/06/04 01:24:07 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Support for 'Compiled Rules'.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../lib/ergo.h"


%% begin_compiled_rule(+Call, +Patt, +OptionList)
%%
%%   Declares a new compiled rule group that will be used
%%   to generate theorems of the form Patt.
%%
%%   The outermost functor of Call is the name of the group.
%%   (It is necessary for the name of the group to be an atom so that
%%    we can generate a unique functor name for the asserted clauses.)
%%
%%   Any members of OptionList must be of the following form:
%%      precondition=Code

begin_compiled_rule(Call, Form, Options) :-
    Call =.. [Name|Args],
    available(Name),
    check_term(Form),
    !,
    %% check that Form does not contain quantifiers or object
    %% variables and that all its meta-variables appear in Call.
    check_or_error(
        collect_obvars(Form, []),
        ['quantifiers/object-vars not allowed in: ',thy(Form)]),
    check_or_error(
        ( collect_metavars(Form, Vars),
          collect_metavars(Call, CallVars),
          subset_var(Vars, CallVars)
        ),
        ['all variables in: ', thy(Form),nl,
         'must also occur in: ', code(Call)]),
    !,
    %% Construct the internal name: cr__(theory)_(name)
    name(Name, NameStr),
    global_state_lookup(current_dir,Theory),
    name(Theory, TheoryStr),
    append("cr__", TheoryStr, S1),
    append(S1, "_", S2),
    append(S2, NameStr, S3),
    name(IName, S3),
    %% Construct the internal arguments and call.
    %%
    %% First argument = tracing information
    %% Last argument =
    append([_|Args], [_], Args2),
    InternalCall =.. [IName|Args2],
    !,
    %% Separate out the precondition.
    (
        delete(precondition = PreCond, Options, Options2)
    ;
        ( PreCond = true,
          Options2 = Options
        )
    ),
    check_or_error(
        Options2 = [], 
        ['unknown options: ', w(Options2)]),
    %% assert this compiled rule into the current theory.
    functor(InternalCall, IName, IArity),
    assert_theory(save_clauses(IName, IArity, IArity)),
    assert_theory(compiled_rule(Name, Call, Form, _, open,
                                PreCond, InternalCall, _)).


%% end_compiled_rule(+Name).
%%
%%  his command completes the construction of the compiled rule 
%%  group called Name.  Once this command has been executed, no
%%  more rules can be added to the group.
%%
%%  Eventually, this command may cause the compiled rule group
%%  to *really* be compiled, rather than just asserted.
%%
end_compiled_rule(Name) :-
    Fact0 = compiled_rule(Name, Call, Form, Options, open, Pre, ICall, Vars),
    query_ancestors(Fact0),
    check_or_error(
        global_state_lookup(current_dir,Thy),
        ['begin/end_compiled_rule must be in the same theory']),
    retract_theory(Fact0, Thy),
    Fact = compiled_rule(Name, Call, Form, Options, closed, Pre, ICall, Vars),
    assert_theory(Fact, Thy).



%% get_compiled_rule(?Thy, ?Name, ?Call, ?Form)
%%
%%  This procedure gets the most commonly needed information about
%%  a compiled rule from the database, without evaluating the rule.
%%
get_compiled_rule(Thy, Name, Call, Form) :-
    query_ancestors(compiled_rule(Name, Call, Form, _, _, _, _, _), Thy).



%% eval_compiled_rule(?Thy, ?Name, +Call, ?InForm, -FinalForm, -Deps, -UDeps).
%% eval_compiled_rule(?Thy, ?Name, ?Call, +InForm, -FinalForm, -Deps, -UDeps).
%% 
%%  This procedure attempts to generate a theorem
%%  that matches the specified arguments of Call and/or Form.
%%  If it succeeds, it instantiates Call and Form as well as
%%  returning the whole theorem in FinalForm.
%%  Deps is a list of dependency variables, to indicate which
%%  conjectures this compiled rule depends upon.
%%
eval_compiled_rule(Thy, Name, Call, InForm, FinalForm, Deps, UseDeps, Trace) :-
    query_ancestors(compiled_rule(Name,Call,Form,_,_,Pre,ICall,_),
                    Thy),
    %% Deduce Form from InForm or FinalForm.
    ( nonvar(InForm) ->
        Form = InForm
    ; compound(FinalForm), functor(FinalForm, F, 2), F == ERGO_IMPLIES ->
        FinalForm = ERGO_IMPLIES(Conds,Form),
        CList0 = [_|_]
    ;
        FinalForm = Form
    ),
    once(Pre),
    % precalculate the expected list of conditions if possible
    (nonvar(Conds) ->
        once(conjunction_to_list(Conds, CList0))
    ;
        true
    ),
    ip_set(cr_dependencies, []),
    ip_set(cr_use, []),
    ip_set(cr_conditions, CList0),
    !,
    arg(1, ICall, Trace),
    call_predicate(ICall),
    ip_lookup(cr_dependencies, Deps),
    ip_lookup(cr_use, UseDeps),
    ip_lookup(cr_conditions, CList1),
    once(append(CList1, [], CList2)),
    % this may be either generating Conds from CList2, or
    % just checking CList2 against an instantiated Conds.
    list_to_conjunction(CList2, Conds),
    choose_conditional(_, FinalForm, Conds, Form).


%% compiled_rule(+Thm, +Name)
%% compiled_rule(+Thm, +(Call :- Code))
%% compiled_rule(+Thm, ?Dir, +Name)
%% compiled_rule(+Thm, ?Dir, +(Call :- Code))
%% compiled_rule(+Thm, ?Dir, +Name, ?(Call :- Code))
%%
%%   Compiles the theorem/axiom called Thm into a Qu-Prolog
%%   clause that becomes part of the definition of the compiled rule
%%   group called Name.  The resulting clause is asserted after
%%   the clauses generated by all previous compiled_rule commands.
%%
%%   Thm must be an axiom or a theorem, not a conjecture/hypothesis
%%   or another compiled rule (because the generation of that 
%%   compiled rule may use conjectures/hypotheses if this command
%%   is called from within a proof).
%%
%%   The second form allows the compiled clause to be specified more
%%   precisely.  Call must match the call term specified in the
%%   declaration of a compiled rule group.  The parameters within
%%   Call may be instantiated.  Code must be a comma-separated sequence
%%   of terms that specify how the conditions and body of Thm are
%%   to be discharged.  The allowable forms for these terms include:
%%   (where N is the number of some condition of Thm, starting from 1)
%%      body       -- 
%%      N=is_hyp   -- discharge condition N using is_hyp(_, _).
%%      N=return   -- discharge condition N by returning it as a
%%                      condition of the resulting theorem.
%%      N=Call2    -- discharge condition N by a recursive call
%%                    to some compiled rule group (not nec. this one).
%%      N          -- an abbreviation of N=Call2, where Call2 is a
%%                      call to *this* compiled rule group.
%%      N=Thm      -- discharge condition N at compile time, using the
%%                    rule called Thm.  Thm must be a fully qualified
%%                    name of the form Thy:Name and the rule that it refers
%%                    must not have any conditions, but may have a body.
%%      N=is_theorem -- discharge condition N by searching (at eval time)
%%                    for a theorem that matches condition N.
%%      {Term}     -- check that Term succeeds.
%%                      (generates code of the form:  \+ \+ Term)
%%
%%      various safe Qu-Prolog/Ergo predicates, such as
%%          var(_), nonvar(_), !, incomplete_retry_delays...
%%          (see gen_code/3 for details of what is allowed).
%%
compiled_rule(Thm, OtherArgument) :-
    compiled_rule(Thm, _, OtherArgument).

compiled_rule(Thm, Dir, Name) :-
    atom(Name),
    !,
    compiled_rule(Thm, Dir, Name, _).
compiled_rule(Thm, Dir, (Call :- Code)) :-
    functor(Call, Name, _),
    !,
    compiled_rule(Thm, Dir, Name, (Call :- Code)).


compiled_rule(Thm, Dir, Name, (Call :- Code)) :-
    query_ancestors(compiled_rule(Name,Call,Form,_,Open,_,ICall,_),
                    Theory),
    check_or_error(
        Open == open,
        ['compiled_rule ', q(Name), ' is closed.']),
    cr_gen_clause(Thm, Dir, Theory:Name, Form, 
                  (Call :- Code),
                  (ICall :- FinalCode)),
    %% ensure that this theory has a 'save_clauses' fact for ICall.
    global_state_lookup(current_dir,CD),
    functor(ICall, IName, IArity),
    (query_theory(save_clauses(IName,IArity,_), CD) ->
        touch_database(CD)
    ;
        assert_theory(save_clauses(IName, IArity, IArity), CD)
    ),
    \+ \+ (thaw_vars((ICall, FinalCode)),
           writelist([q(ICall),' :-',nl,tab,q(FinalCode),'.',nl]),
           assertz((ICall :- FinalCode))).


%% cr_gen_clause(+ThmName, +CRName, ?Dir, +Form, +InClause, ?OutClause).
%%
%%  CRName must be the full name (ie. Thy:Name) of the compiled rule.
%%  Form must be the skeleton (pattern) of that compiled rule
%%  (and must have been instantiated to correspond to Call).
%%  InClause is the given (Call :- BodySpec).
%%  OutClause is the generated (InternalCall :- BodyCode).
%%
cr_gen_clause(Thm, Dir, Theory:Name, Form, (_Call:-Code0), (ICall:-Code)) :-
    %% determine what conditions must be discharged.
    find_rule(Thm, ThmForm, Body),
    choose_conditional(CList, ThmForm, Conds, ThmConcl),
    !,
    best_direction(Dir, Form, ThmConcl),
    choose_direction(Dir, ThmConcl, F, TEq, G),
    choose_direction(FDir,   Form,  F, Eq, G),
    member(FDir, [f, t]),
    is_stronger_than(TEq, Eq),
    incomplete_retry_delays,
    !,
    make_cond_list(Conds, 1, [], CList, TrList),
    make_trace_term(Thm, Dir, TrList, TraceTerm),
    arg(1, ICall, TraceTerm),
    check_term(Form),
    !,
    ((Body == true ->
        CList2 = CList)
    ;
        (CList2 = [(body,Body,body) | CList])
    ),
    %% generate Code from Code0
    gen_code_seq(Code0, CList2, info(Name), Code1),
    incomplete_retry_delays,
    !,
    %% mark this clause as belonging to the current theory.
    global_state_lookup(current_dir,CD),
    functor(ICall, _IName, IArity),
    arg(IArity, ICall, CD),
    %% add code to disable the clause if the compiled rule
    %% is called from a theory above CD.
    ((CD == Theory ->
        Code = Code1)
    ;
        (Code = (ancestor(CD),Code1))
    ).


%% best_direction(?Dir, +Form, +ThmForm)
%%
%% Tweak the default direction for application of a rule by looking at 
%% the forms of the theorem and compiled rule.
%%
%% The final clause ensures that all directions are tried, but 't' may be
%% tried twice.
%%
best_direction(Dir, _F, _TF) :-
    nonvar(Dir),
    !.
best_direction(t, Form, _TF) :-
    functor(Form, F, _),
    \+ is_preorder(F).
best_direction(_Dir, _F, _TF).


%% make_cond_list(+Conjuncts, +Num, +List, ?CondList, ?TraceList)
%%
%%  This turns the antecedents of the rule into a list of
%%  (Path, Term, Trace) triplets.  Also returns a list of Traces.
%%
%%  Eg.  
%%
%%    make_cond_list(A and B, 1, [2], [([2,1],A,AT), ([2,2],B,BT)], [AT,BT]).
%%
make_cond_list(ERGO_TRUE,  1, _, [], []).
make_cond_list(Conds, N, NL, [(RL,C,T)|CList], [T|TList]) :-
    nonvar(Conds),
    Conds = ERGO_AND(C,CS),
    !,
    N1 is N + 1,
    reverse([N|NL], RL),
    make_cond_list(CS, N1, NL, CList, TList).
make_cond_list(C, N, NL, [(RL,C,T)], [T]) :-
    %% C is not a conjunction.
    reverse([N|NL], RL).


%% make_trace_term(+Theorem, +Dir, +TraceList, -TraceTerm)
%%
%% Introduce some abbreviations for common programmatic rule usages
%%
%% Traces must be lists.
%%
%% As normal invokations are traced by lists with at least two elements
%% we use singleton lists for abbreviations.
%%
make_trace_term(r:refleximp, t, [T], [T]) :- !.
make_trace_term(r:refleximp, t, T, [T]) :- !.
make_trace_term(r:transeql, _, [A,B], [(A,B)]) :- !.
make_trace_term(r:transeqv, _, [A,B], [(A,B)]) :- !.
make_trace_term(Thm, Dir, TrList, [Thm|TrList]) :-
    write((Thm, Dir, TrList, TrList)), nl.



%% gen_code_seq(+CodeSpec, +CondList, +Info, -Code).
%%
%%  This turns the code sequence specified by the user (CodeSpec)
%%  into executable Qu-Prolog code.  It ensures that the resulting
%%  code will fulfill the following soundness requirements whenever
%%  it executes without failing:
%%
%%      SR1: the generated code will discharge all conditions in CondList.
%%           (however, it is allowed to discharge them by adding them
%%           to the implicit parameter list 'cr_conditions' at runtime.
%%           They will be returned as conditions of the resulting theorem).
%%
%%      SR2: the generated code must collect the dependency information
%%           of any hypotheses/conjectures that it uses, into the
%%           implicit parameter list 'cr_dependencies'.
%%
%%      SR3: whenever the generated code instantiates a variable that 
%%           may appear in the final theorem, the instantiated term 
%%           must be legal within the theory that this compiled_rule/2
%%           command was executed from.
%%
%%      SR4: the generated code will not have any side effects on the 
%%           current proof, or on any implicit parameters (except for
%%           cr_conditions and cr_dependencies, as mentioned above).
%%           This implies that it cannot call arbitrary
%%           Prolog predicates (eg. trans, h_trans, conjecture).
%%           (NB. it does not check for side effects on the Prolog
%%            database, so we must assume that CodeSpec will not
%%            assert/retract clauses.  A Prolog module system
%%            is needed to fix this properly).
%%
gen_code_seq(Spec, CList, Info, Code) :-
    var(Spec),
    !,
    gen_all_discharge_code(CList, Info, Code).
gen_code_seq((Spec,Rest), CList, Info, (Code,CodeRest)) :-
    !,
    gen_code_frag(Spec, CList, Info, CList2, Code),
    !,
    gen_code_seq(Rest, CList2, Info, CodeRest).
gen_code_seq(Spec, CList, Info, Code) :-
    %% nonvar(Spec), Spec \= (_,_)
    !,
    gen_code_frag(Spec, CList, Info, CList2, Code),
    check_or_error(
        CList2 = [],
        ['the following conditions were not discharged: ',nl,
         tab,thy(CList2)]).


%% gen_all_discharge_code(+CondList, +Info, -Code)
%%
%%  Generates Code to discharge all the remaining conditions in CondList.
%%
gen_all_discharge_code([], _, true) :-
    !.
gen_all_discharge_code([(N,C,T)], Info, Code) :-
    !,
    gen_discharge_code(N, C, Info, _, Code, [], T).
gen_all_discharge_code([(N,C,T) | CList], Info, (Code,Rest)) :-
    gen_discharge_code(N, C, Info, _, Code, [], T),
    gen_all_discharge_code(CList, Info, Rest).


%% gen_code_frag(+Spec, +CList0, +Info, -CList, -Code).
%%
%%  Generates Code from Spec.  If Spec specifies code to
%%  discharge a condition, that condition is removed from
%%  CList0 to give CList, otherwise CList0 and CList are
%%  identical.
%%
gen_code_frag(N, CList0, Info, CList, Code) :-
    (
        N = (Num0=Spec)
    ;
        (Num0 = N)
    ),
    (integer(Num0) ->
        Num = [Num0]
    ;
        (
            list(Num0)
        ;
            Num0 == body
        ),
        Num = Num0
    ),
    !,
    check_or_error(
        delete((Num,Cond,Trace), CList0, CList1),
        [w(Num),' = ',thy(Cond), ' discharged too many times.']),
    gen_discharge_code(Num, Cond, Info, Spec, Code, NewConds, Trace),
    append(NewConds, CList1, CList).
gen_code_frag(Spec, CList0, Info, CList0, Code) :-
    gen_code(Spec, Info, Code).



%% gen_discharge_code(+N, +Cond, +Info, ?Spec, -Code, ?NewConds, ?Trace).
%%
%%  Generates Code to discharge Cond, where Cond is the N'th
%%  condition of the rule being compiled (or if N == body,
%%  then Cond is the Qu-Prolog side condition of that rule).
%%  Spec determines how the condition should be discharged.
%%
gen_discharge_code(body, Code, _Info, _, Code, [], _) :- !.
gen_discharge_code(_N, Cond, info(Name), Spec, ICall, [], NewTr) :-
    var(Spec),
    !,
    %% recursive call to this compiled rule.
    query_ancestor(compiled_rule(Name, _, Cond, _, _, _, ICall, _)),
    incomplete_retry_delays,
    arg(1, ICall, Tr),
    NewTr =.. [Name | Tr],
    !.
gen_discharge_code(_N,Cond,_I, is_hyp,    cr_is_hyp(HN, Cond), [],HN) :-
    !.
gen_discharge_code(_N,Cond,_I, return,    cr_return(RN, Cond), [],r(RN)) :-
    !.
gen_discharge_code(_N,Cond,_I, is_theorem,is_theorem(T:R,Cond),[],T:R) :-
    !.
gen_discharge_code(_N,Cond,_I, is_theorem(T:R),is_theorem(T:R,Cond),[],T:R) :-
    !.
gen_discharge_code( N,Cond,_I, Thy:Name,  Body,   NewConds, NewTr) :-
    !,
    check_or_error(
        find_rule2(Thy, Name, ThmForm, [], Body),
        ['rule ', q(Thy),':', q(Name), ' not found.']),
    check_or_error(
        ( choose_conditional(NewConds, ThmForm, ThmCond, ThmConcl),
          discharges(ThmConcl, Cond)
        ),
        [q(Thy),':',q(Name),' does not discharge condition ',
         q(N),'=',q(Cond)]),
    make_cond_list(ThmCond, 1, N, NewConds, Tr),
    !,
    make_trace_term(Thy:Name, t, Tr, NewTr),
    !.
gen_discharge_code(_N, Cond, _Info, Call, ICall, [], NewTr) :-
    functor(Call, Name, _),
    query_ancestor(compiled_rule(Name, Call, Cond, _, _, _, ICall, _)),
    incomplete_retry_delays,
    arg(1, ICall, Tr),
    NewTr =.. [Name, Tr],
    !.
gen_discharge_code(N, Cond, _Info, Spec, _, _, _) :-
    writelist(['Error: condition number ', w(N), ': ', q(Cond), nl,
               '       cannot be discharged by: ', q(Spec), nl]),
    fail.


%% discharges(+T, +T1)
%%
%%      Unfortunate workaround; T => T1
%%
discharges(T, T1) :-
    choose_direction(_, T, F, E, G),
    choose_direction(_, T1, F, E1, G),
    is_stronger_than(E, E1),
    !.


%% cr_is_hyp(-N, +Form)
%%
%%   Auxillary function for discharging a condition, Form, by
%%   finding a matching hypothesis (and recording its dependencies).
%%
cr_is_hyp(N, Form) :-
    ip_lookup(hyplist, HL),
    member(hrec(N, Form, Del, _, Dep, UDep), HL),
    var(Del),
    incomplete_retry_delays,
    (
        Dep == [] 
    ;
        ip_lookup(cr_dependencies, DepList0),
        collect_vars((Dep, DepList0), DepList),
        ip_set(cr_dependencies, DepList)
    ),
    !,
    ip_lookup(cr_use, CR_UDep),
    ip_set(cr_udep, (UDep, CR_UDep)).


%% cr_return(-N, +Cond)
%%
%%   Auxillary function for discharging a condition, Cond, by
%%   returning it as a condition of the resulting theorem.
%%         
cr_return(N, Cond) :-
    ip_lookup(cr_conditions, CList),
    nth1(N, CList, Cond),
    !.
%%   TO DO: cr_return should perhaps use member_var to match against
%%         existing members of cr_conditions.  However, it must
%%         still be able to instantiate/extend cr_conditions in
%%         order to add new conditions.
%% Ray's final comment on this: obsolete?


%% gen_code(+Spec, +Info, -Code).
%%
%%   Generates Code from Spec.  This procedure handles all Spec's 
%%   that are not discharging conditions.
%%
%%   All code generated by this procedure must satisfy the 
%%   soundness constraints, SR3 and SR4, mentioned above.
%%
gen_code({T},                     _, (\+ \+ T)) :- !.
gen_code(\+ T,                    _, \+ T) :- !.
gen_code(!,                       _, !) :- !.
gen_code(A==B,                    _, A==B) :- !.
gen_code(A\==B,                   _, A\==B) :- !.
gen_code(true,                    _, true) :- !.
gen_code(fail,                    _, fail) :- !.
gen_code(var(A),                  _, var(A)) :- !.
gen_code(nonvar(A),               _, nonvar(A)) :- !.
gen_code(obvar(A),                _, obvar(A)) :- !.
gen_code(integer(A),              _, integer(A)) :- \+ compound(A),!.
gen_code(cr_freeze(A),            _, true) :- freeze(A), !.
gen_code(atom(A),                 _, atom(A)) :- !.
gen_code(atomic(A),               _, atomic(A)) :- !.
gen_code(list(A),                 _, list(A)) :- !.
gen_code(frozen_var(A),           _, frozen_var(A)) :- !.
gen_code(msg(M),                  _, msg(M)) :- !.   
gen_code(less_verbose(msg(M)),    _, less_verbose(msg(M))) :- !.
gen_code(less_verbose(N,msg(M)),  _, less_verbose(N,msg(M))) :- !.
gen_code(incomplete_retry_delays, _, incomplete_retry_delays) :- !.
gen_code(Spec, _, _) :-
    writelist(['Error: unknown code specification: ',q(Spec),nl]),
    fail.

