%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: hook.ql,v 1.29 2003/03/23 09:24:19 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file implements the hook (or call-back) facilities of Ergo, which
%% allow procedures outside the core prover to be called when various
%% events occur.  Tactic files may define additional kinds of events.
%%
%% Events are represented by nonvar terms, possibly containing variables.
%% This allows the procedures attached to a given kind of event to
%% use pattern matching to select some subset of events that they
%% are interested in.  If the event contains variables, they may also
%% instantiate those variables to return information to the initiator
%% of the event.
%%
%% There are two different kinds of events supported: 
%%
%%   1) Announcements.  The procedures attached to these kinds of
%%      events are called `listeners'.  An announcement is broadcast
%%      to *all* listeners that are attached to the announcement,
%%      regardless of whether or not they fail or succeed.  It is also
%%      possible to define a listener that listens to *all* kinds of
%%      announcements (by specifying a variable as the pattern that
%%      matches announcements).
%%
%%   2) Requests.  The procedures attached to these kinds of events
%%      are called `handlers'.  The handlers attached to a particular
%%      kind of request (determined by its outermost functor) are
%%      stored in a list, with more recently added handlers nearer
%%      the front of the list.  A request is passed to each handler
%%      in the associated list, starting from the front, until one
%%      of the handlers succeeds, then no more handlers are called.
%%      If none of the handlers in the list succeed, then some default
%%      action is taken.  Unlike listeners, each handler must be attached
%%      to a specific kind of request, so its request-matching pattern
%%      must have an outermost functor that corresponds to a defined
%%      kind of request.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% add_announcement(Event, Doc).
%%
%%  Declares a new class of announcement that handlers can be attached to.
%%
%%  Event must be an atomic or compound term.  Its outermost functor
%%  must not already be declared as an announcement.
%%  Doc must be a writelist that documents the possible
%%  instantiations and meanings of Event.
%%
%%  mode add_announcement(@compound, @list(nonvar)).
%%  mode add_announcement(@atom,     @list(nonvar)).
%%
add_announcement(Event, Doc) :-
    PRE(add_announcement, std_nonvar(Event)),
    PRE(add_announcement, list(Doc)),
    functor(Event, F, A),
    functor(Patt, F, A),
    check_or_error(
        \+ query_global(announcement(Patt,_)),
        ['announcement ', q(F), '/', q(A), ' is already defined.']),
    assert_global(announcement(Event, Doc)).


%% add_listener(Event, Listener).
%%
%%  Adds Listener (which can be arbitrary Prolog code) to the list of
%%  listeners which will be called whenever Event happens.
%%  Event must be a variable (which means that Listener will receive
%%  all kinds of announcements) or an atomic or compound term whose outermost
%%  functor is declared as an announcement.
%%
%%  mode add_listener(@var,      @code).
%%  mode add_listener(@compound, @code).
%%  mode add_listener(@atom,     @code).
%%
add_listener(Event, Listener) :-
    nonvar(Listener),
    (var(Event) ->
        true
    ;
        check_or_error(
            ( std_nonvar(Event),
              query_global(announcement(Event,_))),
            ['attempt to add listener to unknown announcement: ', q(Event)])
    ),
    assert_global(listener(Event, Listener)),
    hook_immediate(Event, Listener).

%% private hook_immediate(Event, Listener).
%%  Calls Listener immediately if Event is prover(startup) and
%%  the prover has already been started.
%%
hook_immediate(prover(startup), Handler) :-
    loaded_theory(global,_),
    debug(announce_event, ['calling listener: ',q(Handler),
                           ' (for past event).']),
    call_predicate(Handler),
    debug(announce_event, ['exited listener.']),
    !.
hook_immediate(_, _).


%% delete_listeners(Event, Listener).
%%
%%  Deletes all listeners that match Event and Listener.
%%  To avoid deleting too many listeners, both arguments
%%  (especially Listener) should be specified as precisely as possible.
%%
%%  mode delete_listeners(@term, ?code).
%%
delete_listeners(Event, Listener) :-
    retractall_global(listener(Event, Listener)).


%% announce_event(Event).
%%
%%  Calls all listeners associated with the announcement Event.
%%  This always succeeds, regardless of whether or not there are
%%  any listeners or whether or not they fail.  However, listeners
%%  that succeed may return information by instantiating arguments
%%  of Event, if such arguments are provided.
%%
%%  mode announce_event(+compound).
%%  mode announce_event(+atom).
%%
announce_event(Event) :-
    PRE(announce_event, std_nonvar(Event)),
    PRE(announce_event, query_global(announcement(Event, _))),
    functor(Event, Fun, Ar),  % cut down the matching
    functor(E, Fun, Ar),
    findall(E-L,
            ( query_global(listener(E,L)),
              \+ \+ E = Event),
            Listeners),
    debug(announce, ['Listeners: ',q(Listeners)]),
    announce_event2(Listeners, Event).

announce_event2([],_).
announce_event2([E-L|LS],Event) :-
    debug(announce, ['Before: ',q(E = Event)]),
    E = Event,
    % Note that we *want* the above unification to fail if
    % some previous listener has instantiated Event differently
    % to how L wants to.  That is, the first instantiator wins.
    debug(announce, ['After: ',q(E = Event)]),
    call_predicate(L),
    !,
    debug(announce, ['After call: ',q(E = Event)]),
    announce_event2(LS,Event).
announce_event2([_|LS],Event) :-
    announce_event2(LS,Event).

%% TO DO: Reimplement announce_event by storing the listeners for each event
%% in a separate list. This is an efficiency only issue.


%% add_request(Event, Doc).
%%
%%  Declares a new class of request that handlers can be attached to.
%%
%%  Event must be an atomic or compound term.  Its outermost functor
%%  must not already be declared as an request.
%%  Doc must be a writelist that documents the possible
%%  instantiations and meanings of Event.
%%
%%  mode add_request(@compound, @list(nonvar)).
%%  mode add_request(@atomic, @list(nonvar)).
%%
add_request(Event, Doc) :-
    PRE(add_request, std_nonvar(Event)),
    PRE(add_request, list(Doc)),
    functor(Event, F, A),
    functor(Patt, F, A),
    %%errornl(add_request_pattern = Patt),
    check_or_error(
        \+ query_global(request(Patt,_)),
        ['request ', q(F), '/', q(A), ' is already defined.']),
    assert_global(request(Event, Doc)).


%% add_handler(Event, Handler).
%%
%%  Adds Handler (which can be arbitrary Prolog code) to the list of
%%  handlers which will be called whenever Event happens.
%%  Event must be an atomic or compound term that unifies with a previously
%%  declared request.
%%
%%  mode add_handler(@compound, @code).
%%  mode add_handler(@atom, @code).
%%
add_handler(Event, Handler) :-
    nonvar(Handler),
    check_or_error(
        ( std_nonvar(Event),
          query_global(request(Event,_))),
        ['attempt to add handler to unknown event: ', q(Event)]),
    assert_global(handler(Event, Handler)).


%% delete_handlers(Event, Handler).
%%
%%  Deletes all handlers that match Event and Handler.
%%  To avoid deleting too many handlers, both arguments
%%  (especially Handler) should be specified as precisely as possible.
%%
%%  mode delete_handlers(@compound, ?code).
%%  mode delete_handlers(@atom,     ?code).
%%
delete_handlers(Event, Handler) :-
    PRE(delete_handlers, std_nonvar(Event)),
    retractall_global(handler(Event, Handler)).


%% handle_event(Event).
%%
%%  Tries each of the handlers for Event until it finds one that
%%  succeeds.  Fails if there are no handlers that succeed.
%%
%%  mode handle_event(+compound).
%%  mode handle_event(+atom).
%%
handle_event(Event) :-
    PRE(handle_event, std_nonvar(Event)),
    PRE(handle_event, query_global(request(Event, _))),
    query_global(handler(Event,Handler)),
    debug(handle_event, ['calling handler: ',q(Handler),'.']),
    call_predicate(Handler),
    debug(handle_event, ['exited handler.']),
    !.



%% handle_error(Msg).
%% handle_error(Class, Msg).
%%
%%  These are specialised forms of handle_event/1 for error(Class, Msg)
%%  events.  They should be used for all such error events.
%%  Note that handle_error(Msg) is equivalent to handle_error(data,Msg).
%%
%%  Class must be one of:
%%
%%      data     -- an error in data supplied by the user or by a tactic.
%%                  The database is not corrupted, so it is possible to
%%                  recover from the error by failing/throwing back to
%%                  some suitable point, such as a dbms or proof interface.
%%
%%      program  -- some procedure failed to meet its specification
%%                  or was called with illegal arguments.
%%
%%      database -- the database has been corrupted in some way that
%%                  is not recoverable.  
%%
%%  If no handlers handle the error, the default action is to fail.
%%
%%  This procedure also detects recursive calls to itself and
%%  uses error/1 to print errors in such cases, then fails
%%  (it should perhaps abort?).
%%
%%  mode handle_error(@list(term)).
%%  mode handle_error(@atom, @list(term)).
%%
handle_error(Msg) :-
    handle_error(data, Msg).

handle_error(Class,Msg) :-
    ip_lookup(handling_error, T),
    T == true,
    !,
    %% this error was generated while handling another error.
    errornl('Error: error generated during error handling!'),
    error(  'Error class =   '),
    errornl( Class ),
    error(  'Error message = '),
    errornl( Msg ),
    fail.
handle_error(Class,Msg0) :-
    increment_error_count,
    ip_set(handling_error, true),
    mk_error_msg(Class, Msg0, Msg),
    (
        handle_event(error(Class,Msg))
    ;
        writelist(user_error, Msg)
    ),
    !,
    fail.


%% private mk_error_msg(Class, Msg0, Msg).
%%
%%  Msg is the same error message as Msg0 (in writelist format),
%%  but with an introductory message (suitable for the given Class
%%  of errors) tacked on the front and a newline tacked on the rear.
%%
%%  mode mk_error_msg(@atom, @writelist, -writelist).
%%
mk_error_msg(data,     Msg0, ['Error: '|Msg]) :-
    append(Msg0, [nl], Msg).
mk_error_msg(database, Msg0, ['Fatal Error: '|Msg]) :-
    append(Msg0,[nl,'DATABASE IS CORRUPT.  PLEASE REBUILD IT.',nl,nl],Msg).
mk_error_msg(program,  Msg0, ['Program Error: '|Msg]) :-
    append(Msg0,[nl,'PLEASE REPORT THIS TO THE ERGO MAINTAINERS.',nl,nl],Msg).



%% handle_illegal_arg(Procedure, ArgNum, ExpectedType).
%% handle_illegal_arg(Procedure, ArgNum, ExpectedType, ActualArgument).
%%
%%  Generates a program error then fails.
%%
%%  mode handle_illegal_arg(@ground,@integer,@atom).
%%  mode handle_illegal_arg(@ground,@integer,@atom,@term).
%%
handle_illegal_arg(Proc, ArgNum, Expected) :-
    handle_error(
        program,
        ['illegal argument# ',w(ArgNum),' in call to ',q(Proc),'.',nl,
         'Expected argument type is: ',w(Expected),'.']).

handle_illegal_arg(Proc, ArgNum, Expected, Actual) :-
    handle_error(
        program,
        ['illegal argument# ',w(ArgNum),' in call to ',q(Proc),'.',nl,
         'Expected argument type is: ',w(Expected),
         '.  Actual argument is: ',q(Actual),'.']).


%% check_or_error(Test, Msg).
%%
%%  This replaces check_or_writelist/2.
%%
%%  It raises a data error (then fails) iff Test fails.
%%  Msg should not include any initial 'Error: ' or trailing nl.
%%
%%  mode check_or_error(@code, @writelist).
%%
#ifdef NO_INLINE
check_or_error(Test, _) :-
    call_predicate(Test),
    !.
check_or_error(_, Msg) :-
    handle_error(data,Msg).
#endif /* NO INLINE */


