%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: theory_data.ql,v 1.84 2002/06/27 05:10:19 itp Exp $

#include "../lib/ergo.h"


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file contains procedures for extending the current theory.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% declare_name(LHS).
%%
%%  Declares the leftmost-innermost identifier of LHS, which
%%  has the same restrictions as the LHS of a definition.
%%  That is, this procedure has the same effect as declare_name/2,
%%  but calculates the arity of the declared name in the
%%  same way that define_name/2 does.
%%  
%%  mode declare_name(@nonvar).
%%
declare_name(LHS) :-
    check_open(Thy),
    defn_lhs(LHS, Name, Arity, _, _, _),
    check_non_internal_atom(Name),
    CHECK(declare_name, check_arity(Arity)),
    internal_name(IName,Thy,Name),
    check_no_name_clash(Thy,IName,Arity),
    assert_theory(symbol(IName, Arity, declared), Thy).



%% declare_name(Name, Arity).
%%
%%  This is the lowest level command used to declare primitive 
%%  (non-parameterized) constants and primitive functions and
%%  quantifiers in the current theory.
%%  (Use declare_param_const/2 to declare parameterized constants).
%%
%%  Name must be an atom and Arity must be a list giving the
%%  arity of Name.  The members of this list define the allowable
%%  structure of terms constructed using Name, from the innermost
%%  term to the outermost.
%%
%%  Examples:    Name    Arity               Example Terms.
%%               a       []                  a.
%%               b       [2]                 b(_,_), b(1,2).
%%               c       [quant([typed])]    (c x:T _), (c [x:Z] x+x).
%%               d       [quant([untyped,
%%                               typed])]    (d [x,y:T] _).
%%               e       [1,2]               e(_)(_,_), e(1)(2,3).
%%               sum     [2,quant([typed])]  (sum(0,10) [x:Z] x*x).
%%
%%  mode declare_name(@atom, @arity).
%%
declare_name(Name, Arity) :-
    check_non_internal_atom(Name),
    check_open(Thy),
    internal_name(IName,Thy,Name),
    check_arity(Arity),
    check_no_name_clash(Thy,IName,Arity),
    assert_theory(symbol(IName,Arity,declared), Thy).

%% private declare_name(Name, Arity, Thy, Kind).
%%
%%  Reserves Name in Thy with Arity abnd Kind.
%%  Name should have no prefix.
%%  Kind can be 'declared' or 'defined'.
%%
%%  mode declare_name(@atom, @arity, @theory, @atom).
%%
declare_name(Name, Arity, Thy, Kind) :-
    check_non_internal_atom(Name),
    check_or_error(query_theory(status(open), Thy),
                   [w(Thy), ' is not an open theory.']),
    check_or_error(member(Kind, [declared,defined]),
                   [w(Kind), ' must be ''declared'' or ''defined''.']),
    internal_name(IName,Thy,Name),
    check_arity(Arity),
    check_no_name_clash(Thy,IName,Arity),
    assert_theory(symbol(IName,Arity,Kind), Thy).


%% declare_param_const(Name, Value, Body).
%%
%%  Declares a parameterized constant in the current theory.
%%  Each parameterized constant is represented in the theory by an
%%  instance of param_const(FullName,Value), where FullName is Name
%%  prefixed with the current theory and Value is an arbitrary
%%  Prolog term.  However, the set of acceptable terms for Value
%%  can be constrained using Body.  Body is Prolog code that should
%%  only succeed when Value is an acceptable parameter for this
%%  parameterised constant.  Typically Body will constrain Value
%%  to be a ground term of some kind.
%%  
%%  Examples: declare_param_const(intconst, I, 0 =< I).
%%            declare_param_const(progvar, V, atom(V)).
%%
%%  mode declare_param_const(@atom, @term, @code).
%%
declare_param_const(Name, Value, Body) :-
    check_open(Thy),
    check_non_internal_atom(Name),
    nonvar(Body),
    internal_name(IName,Thy,Name),
    check_no_name_clash_const(Thy,param_const(IName,Value)),
    assert_theory(symbol(param_const(IName,Value),
                         when(Body),declared), Thy).


%% declare_object_vars(Prefix).
%%
%%  Declares Prefix to be an object variable prefix in the current theory.
%%
%%  mode declare_object_vars(@atom).
%%
declare_object_vars(Prefix) :- 
    check_current_theory(Thy),
    check_atom(Prefix),
    name(Prefix, Str),
    check_or_error(
        \+ (member(C,Str),
            \+ (0'a =< C, C =< 0'z)),
        ['illegal object variable prefix: ',q(Prefix)]),
    check_no_name_clash_obvar(Thy, Prefix),
    assert_theory(object_vars(Prefix), Thy),
    obvar_prefix(Thy, Prefix).



%% add_axiom(Name, AssumpList, Stmt, Body).
%%
%%  Add a user-level axiom to the current theory.
%%  Name must be an atom.
%%  Body may contain only not_free_in checks.
%%
%%  mode add_axiom(@atom, @list(term), @term, @code).
%%
add_axiom(Name, Ass, Stmt, Body) :-
    check_open(Thy),
    check_legal_rule_name(Name),
    check_no_name_clash_rule(Thy, Name),
    check_or_error(
        ( nonvar(Stmt)
        ; Ass=[_|_]
        ),
        ['axiom ',q(Name),' is too general.']),
    check_term(Stmt),
    check_terms(Ass),
    collect_vars((Ass,Stmt), Vars),
    check_axiom_body(Body, Vars),
    assert_axthm(Name, axiom, Ass, Stmt, Body, Thy).


%% add_param_axiom(Name, RealName, AssumpList, Stmt, Body).
%%
%%  Add a parameterized axiom to the current theory.
%%  Parameterized axioms are not intended to be used by typical
%%  Ergo users, but for building in low level facilities such
%%  as accessing the in-built arithmetic of Prolog.
%%
%%  The axiom can be accessed via Name (which must be an atom)
%%  but RealName will be the name stored within proof records.
%%  The outermost functor of RealName must be the same as Name.
%%  Once the axiom has been successfully applied, the arguments
%%  of RealName must contain sufficient information so that the
%%  axiom can be reapplied successfully given just RealName.
%%  Body is unconstrained, but should be designed to ensure the
%%  above property.
%%
%%  mode add_param_axiom(@atom, @compound, @list(term), @term, @nonvar).
%%
add_param_axiom(Name, RealName, Ass, Stmt, Body) :-
    check_open(Thy),
    check_legal_rule_name(Name),
    check_or_error(
        ( (compound(RealName) ; atom(RealName)),
          functor(RealName, Name, _)
        ),
        ['outermost functor of ',q(RealName),' must be ',q(Name),'.']),
    check_no_name_clash_rule(Thy, Name),
    check_or_error(
        ( nonvar(Stmt)
        ; Ass=[_|_]
        ),
        ['axiom ',q(Name),' is too general.']),
    check_term(Stmt),
    check_terms(Ass),
    assert_axthm(Name, axiom(RealName), Ass, Stmt, Body, Thy).


%% define_name(Head,Term).
%%
%%  Adds the axiom Head=Term to the current theory and declares
%%  the innermost name of Head.  See check_definition for the
%%  restrictions on Head and Term.
%%
%%  mode define_name(@nonvar, @nonvar).
%%
define_name(Head, Term) :-
    check_current_theory(Thy),  %% needn't be open.
    check_definition(Head, Term, Name, Arity, NFIs),
    CHECK(define_name, check_arity(Arity)),
    check_non_internal_atom(Name),
    internal_name(IName,Thy,Name),
    check_no_name_clash(Thy,IName,Arity),
    %% Build a copy of Head that contains IName.
    transform_simple_terms(def_rename(Name,IName), Head, IHead),
    assert_theory(symbol(IName, Arity, defined), Thy),
    atom_append('def_', Name, AxName),
    assert_axthm(AxName,defn,[],ERGO_EQUALS(IHead,Term),NFIs, Thy).

def_rename(Name, IName, In, Out) :-
    (In == Name ->
        Out = IName
    ;
        Out = In
    ).


%% check_definition(LHS, RHS, Name, Arity, Constraints).
%%
%%  Checks the following conditions to ensure that every possible
%%  occurence of a defined term can be unfolded back to a unique
%%  term in the original theory.  The resulting definition axiom is
%%
%%    LHS = RHS provided Constraints
%%
%%  and Name (of arity Arity) is the name that is being defined.
%%
%%  The restrictions that are imposed to ensure this are:
%%
%%    1. LHS is a `well-formed-left-hand-side' (see below)
%%         containing at most one quantified term (Q V B).
%%    2. LHS does not contain any repeated metavariables.
%%    3. LHS does not contain any substitutions.
%%    4. RHS is a legal term of the current theory.
%%    5. Each metavariable in RHS also appears in LHS.
%%    6. RHS does not contain any free object variables.
%%    7. Any occurences of B in RHS must be within the scope of V.
%%
%%  Constraints ensures that all object variables in LHS and RHS
%%  are distinct and not_free_in all the metavariables (except that
%%  V may appear within B).
%%
%%
%%  Defn: A term is a `well-formed-left-hand-side' iff it is either:
%%
%%    1. an atom, or
%%    2. a compound term whose functor is a well-formed-left-hand-side
%%         and whose arguments are distinct metavariables, or
%%    3. a quantified term (Q V B), where Q is a well-formed-left-hand-side,
%%         B is a metavariable and V is a closed list of object variables
%%         (possibly with types, in which case the types must all be
%%         metavariables).
%%
%%  Programmers Notes:
%%    1. This procedure allows Name to be any atom, not necessarily
%%       in internal Ergo format ('Thy.Name').
%%    2. This algorithm for checking higher-order quantifier definitions,
%%       and the theory behind the algorithm, are both quite subtle.
%%       Do not change it without consulting John Staples, Peter Robinson,
%%       Ray Nickson and/or Mark Utting.  Formalizing the proof of
%%       soundness for this definition mechanism would be a useful
%%       research paper.
%%
%%  mode check_definition(@nonvar, @term, ?atom,?arity, ?code).
%%
check_definition(LHS, RHS, Name, Arity, Constraints) :-
    %%
    %% `Parse' LHS and check that it is a fully general pattern.
    defn_lhs(LHS, Name, Arity, Args, BndVars, Body),
    !,
    %% NB. Args is a list of distinct metavariables.
    %%     BndVars is a list of distinct obvars ([] for non-quant defns).
    %%
    %% Check that RHS is a legal term in the current theory and
    %% that it does not contain metavariables that are not in LHS.
    %%
    check_term(RHS),
    collect_metavars(RHS, RHSMetaVars),
    diff_list(RHSMetaVars, [Body|Args], BadMetaVars),
    check_or_error(
        BadMetaVars = [],
        ['illegal metavariables ',q(BadMetaVars),
         ' in definition of ',q(Name)]),
    %%
    %% We make all object variables distinct, to avoid RHS being ambiguous.
    collect_obvars((LHS,RHS), ObVars),
    distinct(ObVars),
    %%
    %% Make all object variables not_free_in LHS and RHS.
    %% This checks that RHS has no free object variables, and adds
    %% constraints that ensure that Args and Body cannot be instantiated
    %% to terms whose object variables will be captured in RHS.
    map(defn_not_free_in((LHS,RHS)), [ObVars]),
    ( BndVars = [] ->
        MetaVars = Args
    ;
        MetaVars = [Body|Args],
        %% A quantifier definition, so we must also check that all
        %% occurences of Body in RHS are within the scope of BndVars.
        %% We do this by temporarily instantiating Body to all of
        %% BndVars and checking that each BndVar is not free in ObRHS.
        %%
        debug(definition, ['checking BndVars is_nfi RHS : ',q(BndVars)]),
        check_or_error(
            \+ \+ (Body = BndVars, map(is_not_free_in_rev(RHS), [BndVars])),
            [q(Body),' is not within the scope of ',q(BndVars),' in ',q(RHS)])
    ),
    %%
    %% All checking finished!
    %%
    %% Finally, we collect the not_free_in constraints that are
    %% relevant to LHS and RHS and return them.
    %%
    append(ObVars, MetaVars, AllVars),
    retry_delays,
    collect_constraints(AllVars, Distincts, NFIs, []),
    debug(definition, ['allvars: ',q(AllVars),nl,
                       tab,'nfis: ',q(NFIs),nl,
                       tab,'distincts: ',q(Distincts)]),
    append(NFIs, Distincts, ConstraintList),
    list_to_comma(ConstraintList, Constraints1),
    %% remove duplicated constraints.
    prolog_code_diff(Constraints1, distinct(BndVars), Constraints),
    debug(definition, ['Final constraints: ',code(Constraints)]).
check_definition(LHS, _, _, _, _) :-
    handle_error(['illegal definition LHS: ',thy(LHS)]).


defn_not_free_in(Defn, ObVar) :-
    ObVar not_free_in Defn,
    !.
defn_not_free_in(Defn, ObVar) :-
    handle_error([q(ObVar),' appears free in ',thy(Defn)]).

is_not_free_in_rev(Term, Var) :- Var is_not_free_in Term.


%% defn_lhs(Head, Name, Arity, MetaVars, ObVars, QuantBody).
%%
%%  Splits Head into the various other components.
%%  Head can contain at most one quantifier, but any number of
%%  compound terms.
%%
%%  Name is the leftmost-innermost atom and Arity is its arity.
%%  MetaVars is the list of all metavariables in Head, excluding QuantBody.
%%  ObVars is the list of bound object variables of the outermost quant term.
%%  QuantBody is the body of the outermost quant term
%%    (it is set to [] when ObVars=[]).
%%
%%  NB. It does not care or check whether or not Name is an
%%      internal name.  This is left to the caller.
%%
%%  Includes the following checks:
%%      * all metavariables in Head are distinct.
%%      * all object variables in Head are distinct.
%%      * Head does not contain any substitutions.
%%
%%  Eg.  defn_lhs(
%%           (!!sum(Lo,Hi) [x,x2] Expr)(Arg),
%%           sum,
%%           [2,quant([untyped,untyped]),1],
%%           [Lo,Hi,Arg],
%%           [x,x2],
%%           Expr).
%%
%%  mode defn_lhs(+nonvar,?atom,?arity,?list(var),?list(obvar),?var).
%%  mode defn_lhs(?nonvar,@atom,@arity,?list(var),?list(obvar),?var).
%%
defn_lhs(Term, Name, Arity, MetaVars, ObVars, QuantBody) :-
    ( list(Arity) ->
        reverse(Arity, Arity2),
        defn_lhs2(Term, Name, Arity2, MetaVars, ObVars, QuantBody)
    ;
        defn_lhs2(Term, Name, Arity2, MetaVars, ObVars, QuantBody),
        reverse(Arity2, Arity)
    ),
    ( delete(quant(_), Arity, A2) ->
        %% check that there is at most one quant in Term
        %% (otherwise qinfo would be used twice)
        \+ delete(quant(_), A2, _),
        check_distinct_metavars([QuantBody|MetaVars])
    ;
        %% no quantifier within Term.
        ObVars = [],
        QuantBody = [],
        check_distinct_metavars(MetaVars)
    ).

defn_lhs2(Name, Name, [], [], _, _) :-
    atom(Name).
defn_lhs2(Head, Name, [Len|Arity2], MetaVars, ObVars, Body) :-
    ( var(Head) ->
        integer(Len)
    ;
        compound(Head)
    ),
    functor(Head, Head2, Len),
    defn_lhs2(Head2, Name, Arity2, MVars2, ObVars, Body),
    Head =.. [_|MVars],
    append(MVars2, MVars, MetaVars).
defn_lhs2(Head, Name, [quant(BArity)|Arity2], MetaVars, ObVars, Body) :-
    ( var(Head) ->
        closed_list(BArity)
    ;
        quant(Head)
    ),
    quantify(Head, Head2, V, Body),
    check_or_error(
        \+ sub(Head),
        ['substitutions not allowed in: ',thy(Head)]),
    defn_lhs2(Head2, Name, Arity2, MVars2, ObVars, Body),
    check_bndvar_list(V, BArity, ObVars, TVars),
    append(TVars, MVars2, MetaVars).



general command
  load_tactic_file(File), load_tactic_file(File, Thy)
extends
  tactic_file_cmds
mode
  load_tactic_file(@filename),
  load_tactic_file(@filename, @theory)
doc
  'Load the given File from the current theory (or the ancestor theory',nl,
  'Thy), which should contain tactics that are',nl,
  'related to the current theory.',
  para,
  'File will NOT be loaded automatically in future sessions.',nl,
  'This is done by using the command ',
      xref(add_tactic_file,global,'general:add_tactic_file/2'),'.'.

load_tactic_file(File) :-
    check_current_theory(Thy),
    operate_tactic_file(Thy, load, File).

load_tactic_file(File, Thy) :-
    operate_tactic_file(Thy, load, File).


general command
  add_tactic_file(File), add_tactic_file(File, Thy)
extends
  tactic_file_cmds
mode
  add_tactic_file(@filename), add_tactic_file(@filename, @theory)
doc
  'Load the given File from the current theory (or the ancestor theory',nl,
  'Thy), which should contain tactics that are',nl,
  'related to the current theory.',nl,
  'In future sessions, File will be loaded automatically whenever',nl,
  'this theory is loaded.',
  para,
  'If a tactic file called File is already loaded, this command ',nl,
  'fails with an error message.',nl,
  'To succeed in such a case, use the command ',
      xref(require_tactic_file,global,'general:require_tactic_file/2'),'.',
  para,
  'Loading the tactics in File for this session only can be done using',nl,
  'the command ',xref(load_tactic_file,global,'general:load_tactic_file/2'),
      '.'.

add_tactic_file(File) :-
    check_current_theory(Thy),
    add_tactic_file(File, Thy).

add_tactic_file(File, Thy) :-
    check_or_error(
        \+ query_theory(tactic_file(File), Thy),
        ['tactic file ', q(File), ' is already loaded.']),
    operate_tactic_file(Thy, load, File),
    record_tactic_file(Thy, File).


general command
  reconsult_tactic_file(File), reconsult_tactic_file(File, Thy)
extends
  tactic_file_cmds
mode
  reconsult_tactic_file(@filename), 
  reconsult_tactic_file(@filename, @theory)
doc
  'Reread a previously loaded theory-specific tactic file after',nl,
  'editing it.',
  para,
  'This command only works for tactics which have not been loaded as',nl,
  'compiled files.'.

reconsult_tactic_file(File) :-
    check_current_theory(Thy),
    operate_tactic_file(Thy, reconsult, File).

reconsult_tactic_file(File, Thy) :-
    operate_tactic_file(Thy, reconsult, File).


general command
  require_tactic_file(File), require_tactic_file(File, Thy)
extends
  tactic_file_cmds
mode
  require_tactic_file(@filename),
  require_tactic_file(@filename, @theory)
doc
  'Add File, if it is has not already been added.  File should',nl,
  'contain tactics that are related to the current theory.',nl,
  'In future sessions, File will be added automatically ',nl,
  'whenever this theory is loaded.',
  para,
  'If a tactic file called File is already loaded into the current,',nl,
  'theory, this command succeeds silently.',nl,
  'To fail in such a case, use the command ',
      xref(add_tactic_file,global,'general:add_tactic_file/2'),'.'.

require_tactic_file(File) :-
    check_current_theory(Thy),
    require_tactic_file(File, Thy).

require_tactic_file(File, Thy) :-
    ( concat_atom([Thy, '/', File], FName), loaded(FName,_,_) ->
        %% already loaded
        true
    ; query_theory(tactic_file(File), Thy) ->
        true
    ;
        operate_tactic_file(Thy, load, File),
        record_tactic_file(Thy, File)
    ).

general command
  ensure_loaded_tactic_file(File), ensure_loaded_tactic_file(File, Thy)
extends
  tactic_file_cmds
mode
  ensure_loaded_tactic_file(@filename),
  ensure_loaded_tactic_file(@filename, @theory)
doc
  'Load File, if it is not already loaded.  File should',nl,
  'contain tactics that are related to the current theory.',nl,
  'In future sessions, File will NOT be loaded automatically ',nl,
  'whenever this theory is loaded.',
  para,
  'If a tactic file called File is already loaded into the current,',nl,
  'theory, this command succeeds silently.',nl,
  'To fail in such a case, use the command ',
      xref(load_tactic_file,global,'general:load_tactic_file/2'),'.'.

ensure_loaded_tactic_file(File) :-
    check_current_theory(Thy),
    ensure_loaded_tactic_file(File, Thy).

ensure_loaded_tactic_file(File, Thy) :-
    ( concat_atom([Thy, '/', File], FName), loaded(FName,_,_) ->
        %% already loaded
        true
    ; query_theory(tactic_file(File), Thy) ->
        true
    ;
        operate_tactic_file(Thy, load, File)
    ).



general command
  compile_tactic_file(File),
  compile_tactic_file(File, Thy)
extends
  tactic_file_cmds
mode
  compile_tactic_file(@filename),
  compile_tactic_file(@filename, @theory)
doc
  'Compile the given source File, which should contain tactics that are',nl,
  'related to the current theory.',
  para,
  'If the file has not been previously loaded as a compiled file, it can be',nl,
  'loaded using either ',
      xref(add_tactic_file,global,'general:add_tactic_file/2'),
      ' (for this and future sessions) or',nl,
  xref(load_tactic_file,global,'general:load_tactic_file/2'),
      ' (for this session only),'.

compile_tactic_file(File) :-
    check_current_theory(Thy),
    operate_tactic_file(Thy, compile, File).

compile_tactic_file(File, Thy) :-
    operate_tactic_file(Thy, compile, File).


general command
  recompile_tactic_file(File),
  recompile_tactic_file(File, Thy)
extends
  tactic_file_cmds
mode
  recompile_tactic_file(@filename),
  recompile_tactic_file(@filename, @theory)
doc
  'Recompile the given source File (which should contain tactics that are',nl,
  'related to the current theory) unless a corresponding object file exists',nl,
  'and is up to date.',
  para,
  'If the file has not been previously loaded as a compiled file, it can be',nl,
  'loaded using either ',
      xref(add_tactic_file,global,'general:add_tactic_file/2'),
      ' (for this and future sessions) or',nl,
  xref(load_tactic_file,global,'general:load_tactic_file/2'),
      ' (for this session only),'.

?- dynamic(suppress_recompile_tactic_file/0).

recompile_tactic_file(_File) :-
    suppress_recompile_tactic_file,
    !.
recompile_tactic_file(File) :-
    check_current_theory(Thy),
    operate_tactic_file(Thy, recompile, File).

recompile_tactic_file(_File, _Thy) :-
    suppress_recompile_tactic_file,
    !.
recompile_tactic_file(File, Thy) :-
    operate_tactic_file(Thy, recompile, File).


general command
  recompile_all_tactic_files
extends
  tactic_file_cmds
mode
  recompile_all_tactic_files
doc
  'Recompile all tactic files in the current theory.',nl,
  'Each file is compiled unless a corresponding object file exists',nl,
  'and is up to date.',
  nl,
  'Recompiling an individual tactic file can be done using the command',nl,
  xref(recompile_tactic_file,global,'general:recompile_tactic_file/2'),'.'
implementation
  (
      check_current_theory(Thy),
      query_theory(tactic_file(File), Thy),
      operate_tactic_file(Thy, recompile, File),
      fail
  ;
      true
  ).


general command
  load_global_tactic_file(File)
extends
  tactic_file_cmds
mode
  load_global_tactic_file(@filename)
doc
  'For this session only, load the given File which should contain',nl,
  'global tactics (that is, tactics that are not theory-specific).',
  para,
  'The global tactics in File can be loaded for all future sessions using',nl,
  xref(add_global_tactic_file,global,'general:add_global_tactic_file/1'),'.'.

load_global_tactic_file(File) :-
    operate_tactic_file(global, load, File).


general command
  add_global_tactic_file(File)
extends
  tactic_file_cmds
mode
  add_global_tactic_file(@filename)
doc
  'Load the given File, which should contain global tactics',nl,
  '(that is, tactics that are not theory-specific).',nl,
  'File will be loaded automatically in future sessions.',
  para,
  'If a global tactic file called File is already loaded,',nl,
  'then this command fails with an error message.',
  para,
  'Loading the tactics in File for this session only can be done using',nl,
  'the command ',
      xref(load_global_tactic_file,global,'general:load_global_tactic_file/1'),'.'.

add_global_tactic_file(File) :-
    check_or_error(
        \+ query_global(tactic_file(File)),
        ['global tactic file ', q(File), ' is already loaded.']),
    operate_tactic_file(global, load, File),
    record_tactic_file(global, File).


general command
  reconsult_global_tactic_file(File)
extends
  tactic_file_cmds
mode
  reconsult_global_tactic_file(@filename)
doc
  'Reread a previously loaded global tactic file',nl,
  '(typically after editing it).',
  para,
  'This command only works for tactics which have not been loaded as',nl,
  'compiled files.'.


reconsult_global_tactic_file(File) :-
    operate_tactic_file(global, reconsult, File).


general command
  require_global_tactic_file(File)
extends
  tactic_file_cmds
mode
  require_global_tactic_file(@filename)
doc
  'Add File, if it has not already been added.  File should contain',nl,
  'global tactics (that is, tactics that are not theory-specific).',nl,
  'File will be loaded automatically in future sessions.',
  para,
  'If a global tactic file called File has already been loaded,',nl,
  'this command succeeds silently.',nl,
  'To fail in such a case, use the command ',
      xref(add_global_tactic_file,global,'general:add_global_tactic_file/1'),
      '.'.

require_global_tactic_file(File) :-
    ( loaded(File,_,_) ->
        %% already loaded
        true
    ; query_global(tactic_file(File)) ->
        %% scheduled to be loaded
        operate_tactic_file(global, load, File)
    ;
        %% not yet known
        operate_tactic_file(global, load, File),
        record_tactic_file(global, File)
    ).

general command
  ensure_loaded_global_tactic_file(File)
extends
  tactic_file_cmds
mode
  ensure_loaded_global_tactic_file(@filename)
doc
  'Load File, if it has not already been loaded.  File should contain',nl,
  'global tactics (that is, tactics that are not theory-specific).',nl,
  'File will be loaded automatically in future sessions.',
  para,
  'If a global tactic file called File has already been loaded,',nl,
  'this command succeeds silently.',nl,
  'To fail in such a case, use the command ',
      xref(load_global_tactic_file,global,'general:load_global_tactic_file/1'),
      '.'.

ensure_loaded_global_tactic_file(File) :-
    ( loaded(File,_,_) ->
        %% already loaded
        true
    ; query_global(tactic_file(File)) ->
        %% scheduled to be loaded
        operate_tactic_file(global, load, File)
    ;
        %% not yet known
        operate_tactic_file(global, load, File)
    ).



general command
  compile_global_tactic_file(File)
extends
  tactic_file_cmds
mode
  compile_global_tactic_file(@filename)
doc
  'Compile the given File, which should contain global tactics',nl,
  '(that is, tactics that are not theory-specific).',
  para,
  'If the compiled file has not previously been loaded, it can be loaded',nl,
  'using ',
      xref(add_global_tactic_file,global,'general:add_global_tactic_file/1'),
      ' (for this and future sessions) or',nl,
  xref(load_global_tactic_file,global,'general:load_global_tactic_file/1'),
      ' (for this session only).'.

compile_global_tactic_file(File) :-
    operate_tactic_file(global, compile, File).


general command
  recompile_global_tactic_file(File)
extends
  tactic_file_cmds
mode
  recompile_global_tactic_file(@filename)
doc
  'Recompile the given source File, unless a corresponding object',nl,
  'file exists and is up to date.',
  para,
  'If the file has not been previously loaded as a compiled file, it can be',nl,
  'loaded using either ',
      xref(add_tactic_file,global,'general:add_tactic_file/2'),
      ' (for this and future sessions) or',nl,
  xref(load_tactic_file,global,'general:load_tactic_file/2'),
      ' (for this session only),'.

recompile_global_tactic_file(File) :-
    operate_tactic_file(global, recompile, File).



%% private operate_tactic_file(Thy, Op, File).
%%
%%  Load/reconsult/compile/recompile/expand a global or
%%  theory-specific tactic file.
%%
%%  mode operate_tactic_file(@theory, @atom, @filename).
%% 
operate_tactic_file(Thy, Op, File) :-
    debug(tactics, [Op, 'ing tactic file ', File, ' in theory ', Thy]),
    check_atom(File),
    ( Thy = global ->
        FName = File,
        Exps = [uniqvars, decl(global), macro]
    ;
        check_or_error(ancestor(Thy),
              [q(Thy), ' is not an ancestor theory']),
        concat_atom([Thy, '/', File], FName),
        ( ergo_version(ergo4) ->
            Exps = [thy(Thy), uniqvars, decl(Thy), macro, prefix(Thy)]
        ;
            Exps = [thy(Thy), uniqvars, decl(Thy), macro]
        )
    ),
    ( Op = load ->
        do_consult(FName, Exps)
    ; Op = reconsult ->
        do_reconsult(FName, Thy, File, Exps)
    ; Op = compile ->
        do_compile(FName, Exps, yes)
    ; Op = recompile ->
        do_compile(FName, Exps, no)
    ; Op = expand ->
        do_expand(FName, Exps)
    ;
        handle_error(program, ['Bad op ', w(Op), ' in operate_tactic_file/3'])
    ),
    debug(tactics, [Op, 'ed tactic file ', File]).

%% private record_tactic_file(Thy, File).
%%
%%  Record the addition of a theory-specific or global tactic file for future
%%  sessions.
%%
%%  mode record_tactic_file(@theory, @filename).
%% 
record_tactic_file(Thy, File) :-
    ( Thy == global ->
        assert_global(tactic_file(File))
    ; 
        assert_theory(tactic_file(File), Thy)
    ).


%% add_preorder(Reln).
%%
%%  This declares Reln as an preorder relation in the current theory.
%%  Reln must be a binary function that is already accessible in the
%%  current theory and there must be axioms/theorems of the form:
%%
%%     Reln(A,A).
%%     Reln(A,B) and Reln(B,C) => Reln(A,C).
%%
%%  mode add_preorder(@atom).
%%
add_preorder(Op) :-
    check_current_theory(Thy),
    check_atom(Op),
    R =.. [Op, A, A],
    check_rule_exists(R),
    T1 =.. [Op, D, E],
    T2 =.. [Op, E, F],
    T3 =.. [Op, D, F],
    check_rule_exists(ERGO_IMPLIES(ERGO_AND(T1,T2),T3)),
    assert_theory(preorder(Op), Thy).


%% add_equivalence(Reln).
%%
%%  The declares Reln as an equivalence relation in the current theory.
%%  Reln must be a binary function that is already accessible in the
%%  current theory.  The add_preorder(Reln) procedure will be called
%%  if Reln is not already a preorder.  In addition, there must be
%%  an axiom/theorem of the form:
%%
%%      Reln(A,B) => Reln(B,A).
%%
%%  NB. This also adds reverse_of(Reln,Reln) to the database, for
%%      simplify_relation/2.  This corresponds to the theorem
%%      Reln(A,B) <=> Reln(B,A), which is equivalent to the above.
%%
%%  mode add_equivalence(@atom).
%%
add_equivalence(Equ) :-
    check_current_theory(Thy),
    check_atom(Equ),
    (
        query_ancestors(preorder(Equ))
    ;
        add_preorder(Equ)
    ),
    !,
    ST1 =.. [Equ, A, B],
    ST2 =.. [Equ, B, A],
    check_rule_exists(ERGO_IMPLIES(ST1,ST2)),
    assert_theory(equivalence(Equ), Thy),
    assert_theory(reverse_of(Equ,Equ), Thy).


%% add_stronger_than(Rel1, Rel2).
%%
%%  Extends the partial order over preorder/equivalence relations
%%  by declaring Rel1 to be a stronger relation than Rel2.
%%  Both relations must have been declared as preorders, and
%%  there must be a rule of the form:
%%
%%       Rel1(A,B) => Rel2(A,B).
%%
%%  NB. Normally this adds the fact stronger_than(Rel1,Rel2) to the
%%      database, but if Rel1 is an equivalence and Rel2 is not, it
%%      also adds (a simplified form of) stronger_than(Rel1,rev(Rel2)).
%%      This is necessary for is_stronger_relation/2.
%%
%%  mode add_stronger_than(@atom, @atom).
%%
add_stronger_than(Op1, Op2) :-
    check_current_theory(Thy),
    check_atom(Op1),
    check_atom(Op2),
    check_preorder(Op1),
    check_preorder(Op2),
    LHS =.. [Op1, A, B],
    RHS =.. [Op2, A, B],
    check_rule_exists(ERGO_IMPLIES(LHS,RHS)),
    assert_theory(stronger_than(Op1, Op2), Thy),
    (is_equivalence(Op1), \+ is_equivalence(Op2) ->
        simplify_relation(rev(Op2), Op2Rev),
        assert_theory(stronger_than(Op1, Op2Rev), Thy)
    ;
        true
    ).


%% add_reverse_of(R1, R2).
%%
%%  This declares R1 to be the reverse relation of R2.
%%  R1 and R2 must both be preorders in the current theory and
%%  there must an axiom/theorem of the form:
%%
%%     R1(A,B) => R2(A,B).
%%
%%  mode add_reverse_of(@atom, @atom).
%%
add_reverse_of(Op1, Op2) :-
    check_current_theory(Thy),
    check_atom(Op1),
    check_atom(Op2),
    check_preorder(Op1),
    check_preorder(Op2),
    check_or_error(
        Op1 \== Op2,
        ['reverse_of(',q(Op1),',',q(Op2),') is not allowed.',nl,
         'Instead, declare ',q(Op1),' as an equivalence relation.']),
    ST1 =.. [Op1, B, C],
    ST2 =.. [Op2, C, B],
    freeze_list([B,C]),
    (
        is_theorem(_, ERGO_EQUIV(ST1,ST2))
    ;
        is_theorem(_, ERGO_EQUALS(ST1,ST2))
    ;
        handle_error(['missing obligation: ',thy(ERGO_EQUIV(ST1,ST2))])
    ),
    !,
    assert_theory(reverse_of(Op2,Op1), Thy),
    assert_theory(reverse_of(Op1,Op2), Thy).


%% check_rule_exists(Stmt).
%%
%%  Succeeds if the current theory or its ancestors contains an
%%  axiom/theorem that can be instantiated to give Stmt,
%%  otherwise it generates an error message.
%%
%%  mode check_rule_exists(@term).
%%
check_rule_exists(A) :-
    \+ \+ ((freeze_vars(A), is_theorem(_, A))),
    !.
check_rule_exists(A) :-
    handle_error(['missing obligation: ', thy(A)]).

