%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: tools.ql,v 1.79 2001/07/19 23:17:10 itp Exp $

#include "../lib/ergo.h"

%% debug(Key, Msg).
%%
%%  If dbg(Key) is true, then Msg is printed, prefixed by 'DEBUG:'
%%  and Key and with a trailing newline added.
%%  
%%  This allows Ergo developers to use assert/retract dbg(Key)
%%  to control which debugging messages they wish to see.
%%  Alternatively, the `-dbg Key' command line argument can be
%%  used to switch on particular groups of debug messages.
%%  (It is best to make Key an atom, so that you can use 
%%  -dbg flags to enable debug messages).
%%
%%  mode debug(@ground, @writelist).
%%
debug(Key, Msg) :-
    (dbg(Key) ->
        writelist_indent(['DEBUG: ',q(Key),tab|Msg],['DEBUG: ...',tab]),
        writelist([nl]),
        flush_output
    ;
        true
    ).


%% pre(Label, Predicate).
%% check(Label, Predicate).
%% post(Label, Predicate).
%%
%%   These predicates always succeed, but generate warning messages
%%   if Predicate fails.  The error message includes Label and Predicate.
%%
%%   Predicate can be any Prolog goal that does
%%   not have side effects (eg. instantiating variables)
%%   Label can be any term, but is typically the name of 
%%   the procedure in which the call to this procedure appears. 
%%
%%   Calling pre, check or post will cause Predicate to be evaluated
%%   in any compilation of Ergo. However, if you wich to only evaluate
%%   the Predicate in test code, then the equivalent
%%
%%   PRE(Label, Predicate)
%%   CHECK(Label, Predicate)
%%   POST(Label, Predicate)
%%
%%   defined in prover/lib/ergo.h should be used instead.
%%
%%   mode pre(  @term, @code).
%%   mode check(@term, @code).
%%   mode post( @term, @code).
%%
pre(_Label, Pred) :-
    \+ \+ call_predicate(Pred),
    !.
pre(Label, Pred) :-
    handle_error(program, ['precondition ',q(Label),
                           ' (', q(Pred), ') failed.']).

check(_Label, Pred) :-
    \+ \+ call_predicate(Pred),
    !.
check(Label, Pred) :-
    handle_error(program, ['assertion ',q(Label),
                           ' (', q(Pred), ') failed.']).

post(_Label, Pred) :-
    \+ \+ call_predicate(Pred),
    !.
post(Label, Pred) :-
    handle_error(program, ['postcondition ',q(Label),
                           ' (', q(Pred), ') failed.']).



%% elem(?E, +List).
%%
%%  Same as member(E, List), but it can also unify E with List
%%  if List is not a list.  That is, elem(a,a) succeeds.
%%
%%  mode elem(?term, ?term).
%%
elem(E, L) :-
    (list(L) ->
        member(E,L)
    ;
        E = L
    ).



%% list_to_comma(List, Code).
%%
%%  Code is a comma-separated sequence of the goals in List.
%%
%%  mode list_to_comma(@list(compound), -compound).
%%
list_to_comma([],     Code) :- !, Code = true.
list_to_comma([C],    Code) :- !, Code = C.
list_to_comma([C|CS], (C,CS2)) :- list_to_comma(CS,CS2).



%% ground(Term).
%%
%%  Succeeds iff Term is ground (contains no object or meta-variables).
%%
%%  mode ground(@ground).
%%
ground(Term) :-
    collect_vars(Term, []).


%% is_legal_mode_list(Modes,Func,Arity).
%% is_legal_mode(Mode,Func,Arity).
%%
%%  Succeeds iff Mode (or each member of the Modes list) is a
%%  legal mode declaration for Func/Arity. 
%%  The outermost functor of each mode specification
%%  must be Func/Arity and any arguments must all be of the form
%%  Op Type, where Op is '@', '+', '-' or '?' and Type is a ground
%%  term that describes the type of that argument upon success.
%%
%%  mode is_legal_mode_list(@list(mode), @atom, @integer).
%%  mode is_legal_mode(@mode, @atom, @integer).
%%
is_legal_mode_list([M],F,A) :-
    !,
    is_legal_mode(M,F,A).
is_legal_mode_list([M|MS],F,A) :-
    %% \+ MS = [].
    A > 0,  %% atoms cannot have multiple modes.
    !,
    is_legal_mode(M,F,A),
    is_legal_mode_list(MS,F,A).

is_legal_mode(arity(Term,_),F,A) :-
    !,
    ground(Term),
    functor(Term,F,A),
    is_legal_mode_args(A,Term).
is_legal_mode(Term,F,A) :-
    ground(Term),
    functor(Term,F,A),
    is_legal_mode_args(A,Term).

is_legal_mode_args(0, _) :-
    !.
is_legal_mode_args(N, Term) :-
    %% N > 0,
    arg(N, Term, Arg),
    mode_io_type(Arg, _, _),
    N1 is N - 1,
    is_legal_mode_args(N1, Term).


%% mode_io_type(ModeArg, IO, Type).
%%
%%  Succeeds iff ModeArg is of the form IO(Type), where IO is one
%%  of the following four input/output specifiers:
%%
%%      @  -- constant input argument (unchanged by the procedure)
%%      +  -- input argument (argument must be instantiated upon entry)
%%      ?  -- input or output argument.
%%      -  -- output-only argument (must be a variable upon entry).
%%
%%  mode mode_io_type(?compound, ?atom, ?term).
%%
mode_io_type(@T, @, T).
mode_io_type(+T, +, T).
mode_io_type(-T, -, T).
mode_io_type(?T, ?, T).


%% first_matching_mode(ModeList, Call, Mode).
%% first_matching_mode(ModeList, Call, Mode, Relax).
%%
%%  ModeList must be a non-empty list of mode declarations.
%%  Mode is the first mode declaration in ModeList that seems
%%  to correctly describe Call, which is a term whose outermost
%%  functor must be the same as the outermost functor of Mode.
%%
%%  Relax is set to relax when satisfies_mode is supposed to 
%%  succeed even when parameters are variables altough this
%%  is not reflexted in the mode declaration.
%%
%%  mode first_matching_mode(@list(mode), @cmd, @mode).
%%  mode first_matching_mode(@list(mode), @cmd, @mode, @nonvar).
%%
first_matching_mode(ModeList, Call, Mode) :-
    first_matching_mode(ModeList, Call, Mode, type_check).
first_matching_mode([M|_], Cmd0, M, Relax) :-
    nonvar(Relax),
    M =.. [F|MArgs],
    Cmd0 =.. [F|CArgs],
    map(satisfies_mode(Relax), [MArgs,CArgs]),
    !.
first_matching_mode([_|MS], Cmd0, M, Relax) :-
    first_matching_mode(MS, Cmd0, M, Relax).


%% satisfies_mode(Relax, ArgMode, Value).
%%
%%  Succeeds if a quick check seems to indicate that Value
%%  satisfies the mode and type given by ArgMode.
%%  ArgMode must be a meta-type (atom, integer, theory, thy(term) etc.)
%%  prefixed by an input-output specifier ('@', '+', '?' or '-').
%%
%%  When Relax is set to relax then type checking is relaxed so that
%%  satisfies_mode succeeds when Val is a var.
%%
%%  satisfies_mode(@nonvar, @gcomp, @term).
%%
satisfies_mode(relax, _, Val) :- 
    var(Val), !.
satisfies_mode(_, '@'(Type), Val) :-
    satisfies_type(Type, Val).
satisfies_mode(_, '-'(_), Val) :-
    var(Val).
satisfies_mode(_, '+'(_), Val) :-
    nonvar(Val).
satisfies_mode(_, '?'(_), Val).


%% satisfies_type(Type, Value).
%%
%%  Succeeds if Value is a member of the Prolog meta-type described
%%  by Type.  However, it only checks a few common/simple types and
%%  assumes that all other values do satisfy the given type, so the
%%  procedure would be more aptly called 'seems_to_satisfy_type'!
%%  (It's main use is as a support routine for first_matching_mode/3).
%%
%%  mode satisfies_type(@ground, @term).
%%
satisfies_type(Type, Value) :-
    query_global(type(Type, Extends, _, Pred)),
    !,
    satisfies_type_pred(Value, Pred),
    map(satisfies_ext_type(Value), [Extends]).
satisfies_type(_, _).

satisfies_type_pred(Value, Pred) :-
    (Pred = notypecheck ->
        true
    ;
        call_predicate(Pred(Value))
    ).

satisfies_ext_type(Value, Type) :-
    satisfies_type(Type, Value).

%% member_var(Element, List).
%%
%%  Succeeds iff Element is a member of List, using == for comparisons.
%%
%%  mode member_var(@term, @list(term)).
%%
member_var(E, L) :- member_eq(E,L).
% member_var(E, [A|B]) :-
%       (E == A ->
%           true
%       ;
%           member_var(E,B)
%       ).


%% insert_list(Value, List0, List).
%%
%%  mode insert_list(@term, @list(term), ?list(term)).
%%
%%  Inserts Value into List0 to give List, unless Value is already
%%  in List0, in which case List=List0.
%%
insert_list(Val, List0, List) :-
    (member_var(Val, List0) ->
        List = List0
    ;
        List = [Val|List0]
    ).



%% subset_var(+L1, +L2)
%%
%%   Succeeds iff L1 is a subset of L2.
%%   L1 and L2 can be unordered and contain duplicates.
%%   Variables within L1 and L2 will *not* be instantiated.
%%
subset_var([], _).
subset_var([L|L1], L2) :-
    member_var(L, L2),
    !,
    subset_var(L1, L2).


%% append( +ListOfLists, ?List ).
%%
%%   Warning: this can go into an infinite loop upon backtracking.
%%
append([], []) :- !.
append([L], L) :- !.
append([L|LS], Out) :-
    append(LS, Tmp),
    append(L, Tmp, Out).



%% nth1(N, List, Val).
%%
%%  Selects the Nth element of List and unifies it with Val.
%%  Note that the numbers start from 1.
%%
%%  mode nth1(?integer, ?list(term), ?term).
%%
nth1(N, L, X) :-
    integer(N),
    !,
    nth1_element(N, L, X).
nth1(N, L, X) :-
    nth1_index(N, L, X).

nth1_element(1, [X|_], X) :-
    !.
nth1_element(N, [_|L], X) :-
    N > 1,
    N1 is N - 1,
    nth1_element(N1, L, X).

nth1_index(1, [X|_], X).
nth1_index(N, [_|L], X) :-
    nth1_index(N1, L, X),
    N is N1 + 1.


%% select(N, Val0, List0, Val, List).
%%
%%  List contains the same elements as List0, except that its
%%  Nth member is Val, whereas the Nth member of List0 is Val0.
%%  Thus, N=1 replaces the head of the list.
%%
%%  mode select(?integer, ?term, ?list(term), ?term, ?list(term)).
%%
select(N, Val0, List0, Val, List) :-
    integer(N),
    !,
    select_element(N, Val0, List0, Val, List).
select(N, Val0, List0, Val, List) :-
    select_index(N, Val0, List0, Val, List).

select_element(1, X, [X|L], Y, [Y|L]).
select_element(N, X, [A|XL], Y, [A|YL]) :-
    N > 1,
    N1 is N - 1,
    select_element(N1, X, XL, Y, YL).

select_index(1, X, [X|L], Y, [Y|L]).
select_index(N, X, [A|XL], Y, [A|YL]) :-
    select_index(N1, X, XL, Y, YL),
    N is N1 + 1.

%% correspond(Val1, List1, List2, Val2).
%%
%%  Val1 is the nth element of List1, and Val2 is the nth element of List2.
%%
%%  mode correspond(?term, ?list(term), ?list(term), ?term).
%% 
correspond(X, [X|_], [Y|_], Y).
correspond(X, [_|L], [_|M], Y) :-
    correspond(X, L, M, Y).

%% atom_append(A, B, AB)
%% atom_string_append(A, B, AB)
%% string_atom_append(A, B, AB)
%%
%%  mode atom_append(@atom, @atom, ?atom).
%%  mode atom_string_append(@atom, @list(integer), ?atom).
%%  mode string_atom_append(@list(integer), @atom, ?atom).
%%
%%  The atom AB is the concatenation of atoms/strings A and B.

atom_append(N1, N2, N3) :-
    atom_concat(N1, N2, N3).

atom_string_append(N1, N2name, N3) :-
    name(N1, N1name),
    append(N1name, N2name, N3name),
    name(N3, N3name).

string_atom_append(N1name,  N2, N3) :-
    name(N2, N2name),
    append(N1name, N2name, N3name),
    name(N3, N3name).



%% atomlist_to_atom(AtomList, Atom).
%%
%%  Eg. Convert [a, b, c] into 'a/b/c'.
%%
%%  mode atomlist_to_atom(@list(atom), ?atom).
%%
atomlist_to_atom(AL, A) :-
    concat_atom(AL, '/', A).



%% collect_allvars(Term, VarList).
%% collect_obvars(Term, VarList).
%% collect_metavars(Term, VarList).
%% collect_thawvars(Term, VarList).
%%
%%  These procedures extract variables out of Term.
%%  The resulting VarList will not contain any duplicates.
%%
%%  collect_allvars/2 collects both object and meta variables.
%%      (this is similar to the Qu-Prolog collect_vars/2
%%       library procedure, except that this version removes
%%       duplicates from the output list).
%%
%%       Note: if you don't mind having duplicates in the resulting
%%       list, then use Qu-Prolog collect_vars/2 instead, because it
%%       is more efficient.
%%
%%  collect_obvars/2 collects object variables only.
%%
%%  collect_metavars/2 collects meta variables only.
%%
%%  collect_thawvars/2 collects thawed meta%object variables only.
%%
%%  mode collect_allvars(+term, ?list(anyvar)).
%%  mode collect_obvars(+term, ?list(obvar)).
%%  mode collect_metavars(+term, ?list(var)).
%%  mode collect_thawvars(+term, ?list(anyvar)).
%%
collect_allvars(Term, Vars)  :-
    collect_vars(Term, Vars).

collect_obvars(Term, Vars)   :-
    collect_vars(Term, Vars1),
    filter(obvar, Vars1, Vars).

collect_metavars(Term, Vars) :-
    collect_vars(Term, Vars1),
    filter(var, Vars1, Vars).
    
collect_thawvars(Term, Vars) :-
    collect_vars(Term, Vars1),
    filter(thawed_var, Vars1, Vars).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   Sets implemented as ordered lists (Quintus compatible).
%%
%%   Operations such as union, intersection, difference are O(N)
%%   on the length of the bigger set.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% NB 1. All comparisons are done using ==.
%%
%% NB 2. As well as the following procedures, you can also use 
%%       member_var/2 and length/2 on these ordered sets.
%%
%% NB 2. Other Quintus ordered set procedures that are not implemented
%%       yet are: 
%%             ord_add_element/3 -- add a given element to a set.
%%             ord_del_element/3 -- delete a given element to a set.
%%             ord_setproduct/3
%%             ord_symdiff/3
%%             ord_intersection/2 -- the intersection of a list of sets.
%%             ord_union/4  -- calculates the union *and* difference.


%% is_ordset(List).
%%
%%  true if List is a closed list and its elements are sorted
%%  in increasing order with no duplicates.
%%
%%  mode is_ordset(@list(term)).
%%
is_ordset(V) :-
    var(V),
    !,
    fail.
is_ordset([]).
is_ordset([_]).
is_ordset([A,B|L]) :-
    A @< B,
    is_ordset([B|L]).


%% list_to_ord_set(List, Set).
%%
%%  This turns a list into an ordered set by sorting it and
%%  removing all duplicates.
%%
%%  mode list_to_ord_set(+list(term), ?list(term)).
%%
list_to_ord_set(L, S) :- sort(L, S).


%% ord_seteq(Set1, Set2).
%%
%%  Succeeds when Set1 and Set2 contain the same elements.
%%
ord_seteq(S1,S2) :-
    S1 == S2.


%% ord_disjoint(Set1, Set2).
%%
%%  Succeeds when Set1 and Set2 have no elements in common.
%%
ord_disjoint(Set1, Set2) :-
    ord_intersection(Set1, Set2, []).


%% ord_intersect(Set1, Set2).
%% ord_intersection(Set1, Set2, Intersection).
%%
%%  ord_intersect succeeds iff Set1 and Set2 have at least one element
%%  in common.  ord_intersection/3 returns the intersection of Set1 and 
%%  Set2 in Intersection.
%%
%%  mode ord_intersect(@list(term), @list(term)).
%%  mode ord_intersection(@list(term), @list(term), ?list(term)).
%%
ord_intersect(S1, S2) :- 
    %% A more efficient implementation might stop as 
    %% soon as a common element is found.
    ord_intersection(S1,S2,[_|_]).

ord_intersection([],     _,      []) :- !.
ord_intersection([_|_],  [],     []) :- !.
ord_intersection([A|AS], [B|BS], Result) :-
    compare(Op,A,B),
    (Op = '=' ->
        Result = [A|Result2],
        ord_intersection(AS,BS,Result2)
    ; Op = '<' ->
        ord_intersection(AS,[B|BS],Result)
    ;
        %% Op = '>'
        ord_intersection([A|AS],BS,Result)
    ).


%% ord_subset(Set1, Set2).
%%
%%  Succeeds when Set1 is a subset of Set2.
%%
%%  mode ord_subset(@list(term), @list(term)).
%%
ord_subset([],     _).
ord_subset([A|AS], [B|BS]) :-
    compare(Op,A,B),
    (Op = '=' ->
        ord_subset(AS,BS)
    ; Op = '<' ->
        fail
    ;
        %% Op = '>'
        ord_subset([A|AS],BS)
    ).



%% ord_subtract(Set1, Set2, Diff).
%%
%%  Succeeds when Diff is the set: Set1 minus Set2.
%%
%%  mode ord_subtract(@list(term), @list(term), ?list(term)).
%%
ord_subtract([],     _,      []) :- !.
ord_subtract(AS,     [],     AS) :- !.
ord_subtract([A|AS], [B|BS], Result) :-
    compare(Op,A,B),
    (Op = '=' ->
        ord_subtract(AS,BS,Result)
    ; Op = '<' ->
        Result = [A|Result2],
        ord_subtract(AS,[B|BS],Result2)
    ;
        %% Op = '>'
        ord_subtract([A|AS],BS,Result)
    ).


%% ord_union(Set1, Set2, Union).
%%
%%  Succeeds when Union is the set: Set1 union Set2.
%%
%%  mode ord_union(@list(term), @list(term), ?list(term)).
%%
ord_union([],     BS,     BS) :- !.
ord_union(AS,     [],     AS) :- !.
ord_union([A|AS], [B|BS], Result) :-
    compare(Op,A,B),
    (Op = '=' ->
        Result = [A|Result2],
        ord_union(AS,BS,Result2)
    ; Op = '<' ->
        Result = [A|Result2],
        ord_union(AS,[B|BS],Result2)
    ;
        %% Op = '>'
        Result = [B|Result2],
        ord_union([A|AS],BS,Result2)
    ).


%% ord_union(Sets, Union).
%%
%%  Sets is a (not necessarily sorted) closed list of sets.
%%  Union is the distributed union of all the sets in Sets.
%%
%%  mode ord_union(@list(list(term)), ?list(term)).
%%
ord_union(Sets, Union) :- ord_union2(Sets, [], Union).

ord_union2([], Union, Union).
ord_union2([S|SS], Sofar, Union) :-
    ord_union(S, Sofar, Sofar2),
    ord_union2(SS, Sofar2, Union).

%% is_chars(Chars).
%%
%%  True iff Chars is a list of characters (a `chars').
%% 
%%  mode is_chars(@term).
%%
is_chars([C|Cs]) :-
    integer(C),
    C >= 0,
    C < 128,
    is_chars(Cs).
is_chars(CS) :-
    atom(CS),
    CS = [].

%% subterm_path(Term, Path, SubTerm).
%%
%%  Finds Subterm at position Path in Term.
%%  Substitutions are evaluated as much as possible, but will be
%%  ignored if they cannot be pushed inside a quantifier due to lack
%%  of information about the distinctness of object variables.
%%  The b(VarList) form is preferred for members of Path that enter
%%  quantifier bodies; similarly the t(Number) form for types.
%% 
%%  mode subterm_path(@term, ?subterm_path, ?term).
%%
subterm_path(Term, [], Term).
subterm_path(Term, [P|Path], ST) :-
    compound(Term),
    functor(Term, F, A),
    F \== param_const,
    between(1, A, P),
    arg(P, Term, Term2),
    subterm_path(Term2, Path, ST).
subterm_path(Term, [P|Path], ST) :-
    quant(Term),
    quantify(Term2, _,VT,Body),
    Term = Term2,
    ( P = b(Vars) ->
        get_bound_obvars(VT, Vars)
    ; P = 2 ->
        true
    ;
        fail
    ),
    subterm_path(Body, Path, ST).
subterm_path(Term, [t(N)|Path], ST) :-
    quant(Term),
    quantify(Term2, _,VT,_),
    Term = Term2,
    nth1(N, VT, _:Type),
    subterm_path(Type, Path, ST).


%% subterm_change_path(Term0, Path, SubTerm0, SubTerm, Term).
%%
%%  Finds SubTerm0 at position Path in Term0, and constructs Term which is like
%%  Term0 except it has SubTerm in place of SubTerm0.
%%
%%  mode subterm_change_path(@term, ?subterm_path, ?term, @term, -term).
%%
subterm_change_path(Term0, [], Term0, Term, Term).
subterm_change_path(Term0, [P|Path], ST0, ST, Term) :-
    compound(Term0),
    functor(Term0, F, A),
    functor(Term, F, A),
    between(1, A, P),
    P1 is P - 1,
    same_args(Term0, Term, 1, P1),
    P2 is P + 1,
    same_args(Term0, Term, P2, A),
    arg(P, Term0, Term2),
    arg(P, Term, Term3),
    subterm_change_path(Term2, Path, ST0, ST, Term3).
subterm_change_path(Term0, [b(ObVars)|Path], ST0, ST, Term) :-
    quant(Term0),
    quantify(Term2, Quant,VT,Body0),
    Term0 = Term2,
    get_bound_obvars(VT, ObVars),
    quantify(Term, Quant, VT, Body),
    subterm_change_path(Body0, Path, ST0, ST, Body).
subterm_change_path(Term0, [t(N)|Path], ST0, ST, Term) :-
    quant(Term0),
    quantify(Term2,Quant,VT0,Body),
    Term0 = Term2,
    select(N, Obvar:Type0, VT0, Obvar:Type, VT),
    quantify(Term,Quant,VT,Body),
    subterm_change_path(Type0, Path, ST0, ST, Type).


%% get_bound_obvars(BoundVarList, Obvars).
%%
%%  mode get_bound_obvars(@list(term), @list(obvar)).
%%
get_bound_obvars([], []).
get_bound_obvars([V|Rest], [V|VS]) :-
    obvar(V),
    !,
    get_bound_obvars(Rest, VS).
get_bound_obvars([V:_|Rest], [V|VS]) :-
    obvar(V),
    get_bound_obvars(Rest, VS).


%% distinct(ObVarList).
%%
%%  Make the object variables in ObVarList pairwise distinct.
%%
%%  mode distinct(@list(obvar)).
%%
distinct([]).
distinct([V|OV]) :-
    obvar(V),
    V not_free_in OV,
    distinct(OV).


%% call_queries(Queries).
%%
%%  Call each of a list of Prolog queries.
%%
%%  mode call_queries(+list(compound)).
%%
call_queries([]).
call_queries([(?- Q) | Qs]) :-
    call_predicate(Q),
    call_queries(Qs).


%% copy_member(Term, List).
%%
%%  Term unifies with some member of List.  Unlike ordinary member/2 List
%%  is left unchanged.
%%
%%  mode copy_member(?term, @list(term)).
%%
copy_member(Term, List) :-
    member(Term1, List),
    Term ?= Term1,
    copy_term(Term1, Term).

%% ctr_ind(Name, Index).
%%
%%  A (dynamic) table of counter names and indexes.
%% 
?- dynamic(ctr_ind/2).

%% ctr_get(Ctr, Val).
%% ctr_set(Ctr, Val).
%% ctr_inc(Ctr, Val).
%
%%  Get/set/increment a global, persistent counter.
%%
%%  ctr_alloc/2, ctr_get1/2 and ctr_set1/2 are foreign functions, loaded in
%%  unix.ql.
%% 
%%  mode ctr_get(@ground, -integer).
%%  mode ctr_set(@ground, @integer).
%%  mode ctr_inc(@ground, -integer).
%%
ctr_get(Ctr, Val) :-
    ( ctr_ind(Ctr, Ind) ->
        ctr_get1(Ind, Val)
    ;
        ctr_alloc(0, Ind),
        asserta(ctr_ind(Ctr, Ind)),
        Val = 0
    ).

ctr_set(Ctr, Val) :-
    ( ctr_ind(Ctr, Ind) ->
        ctr_set1(Ind, Val)
    ;
        ctr_alloc(Val, Ind),
        asserta(ctr_ind(Ctr, Ind))
    ).

ctr_inc(Ctr, Val) :-
    ( ctr_ind(Ctr, Ind) ->
        ctr_get1(Ind, Val0),
        Val is Val0 + 1,
        ctr_set1(Ind, Val)
    ;
        Val is 1,
        ctr_alloc(Val, Ind),
        asserta(ctr_ind(Ctr, Ind))
    ).

%% save_all_ctrs.
%% restore_all_ctrs.
%%
%%  Save/restore all counters.
%%
save_all_ctrs :-
    retractall(ctr_save(_,_)),
    ctr_ind(Name, Ind),
    ctr_get1(Ind, Val),
    asserta(ctr_save(Name, Val)),
    fail.
save_all_ctrs :-
    retractall(ctr_ind(_, _)).

restore_all_ctrs :-
    retract(ctr_save(Name, Val)),
    ctr_set(Name, Val),
    fail.
restore_all_ctrs.


%% prolog_code_diff(Code1, Code2, Diff).
%%
%%  All three arguments are Qu-Prolog code.
%%  Diff is the set of predicates that are in Code1,
%%  but are not in Code2.
%%
%%  This procedure knows the Qu-Prolog semantics of a few predicates
%%  such as ','/2, not_free_in/2 and distinct/1 and may simplify and 
%%  rearrange such predicates, but it does not handle all Prolog operators.
%%  The code in Diff will be in a simplified form.  For example
%%  it will not contain both x not_free_in y and y not_free_in x.
%%
%%  It uses ==/2 and is careful not to instantiate Code1 or Code2.
%%
%%  There are three main ways of using this:
%%    1) To compute the difference between Code1 and Code2 into Diff.
%%    2) With Diff=true, it checks that Code2 implies Code1.
%%    3) with Code2=true, it acts as a code simplifier, producing
%%       a simplified form of Code1 into Diff.
%%
%%  mode prolog_code_diff(@code, @code, ?code).
%%
prolog_code_diff(Code1, Code2, Diff) :-
    prolog_code_diff(Code1, Code2, true, Diff).

%% private prolog_code_diff(C1, C2, Out0, Out).
%%
%%  Out0 is the output code accumulated so far, and Out is
%%  Out0 plus all predicates that are in C1 and not in C2.
%%
%%  mode prolog_code_diff(@code, @code, @code, ?code).
%%
prolog_code_diff(Code, Given, Out0, Out) :-
    std_nonvar(Code),
    Code = (C1,C2),
    !,
    prolog_code_diff(C2, Given, Out0, Out1),
    prolog_code_diff(C1, Given, Out1, Out).
prolog_code_diff(Code, Given, Out0, Out) :-
    std_nonvar(Code),
    Code = (Obvar not_free_in Term),
    list(Term),
    !,
    ( Term = [T|TS] ->
        %% split Obvar not_free_in [T|TS] into two conjuncts.
        prolog_code_diff(Obvar not_free_in TS, Given, Out0, Out1),
        prolog_code_diff(Obvar not_free_in T, Given, Out1, Out)
    ;
        Out = Out0
    ).
prolog_code_diff(Pred, Given, Out0, Out) :-
    prolog_code_contains(Given, Pred),
    !,
    Out = Out0.
prolog_code_diff(Pred, _, Out0, Out) :-
    prolog_code_contains(Out0, Pred),
    !,
    Out = Out0.
prolog_code_diff(Pred, _, Out0, Out) :-
    %% Pred is not (obviously) in Given or Out0, 
    %% so we add it the output.
    prolog_code_union(Pred, Out0, Out).


%% prolog_code_contains(Code, Pred).
%%
%%  Succeeds iff Pred is one of the predicates in Code.
%%  Pred should be a single predicate (not a ','/2 term)
%%  and if it is a not_free_in/2 predicate, its second argument
%%  should be non-compound.
%%
%%  It uses ==/2 and is careful not to instantiate Pred or Code.
%%
%%  mode prolog_code_contains(@code, @code).
%%
prolog_code_contains(Code, Pred) :-
    Code == Pred,
    !.
prolog_code_contains(_, Pred) :-
    Pred == true,
    !.
prolog_code_contains(Code, Pred) :-
    std_nonvar(Code),
    prolog_code_nonvar_contains(Code, Pred).


prolog_code_nonvar_contains((C1,_), Pred) :-
    prolog_code_contains(C1, Pred),
    !.
prolog_code_nonvar_contains((_,C2), Pred) :-
    prolog_code_contains(C2, Pred),
    !.
prolog_code_nonvar_contains(V not_free_in T, Pred) :-
    std_nonvar(Pred),
    Pred = V2 not_free_in T2,
    !,
    prolog_code_contains_nfi(V, T, V2, T2).
prolog_code_nonvar_contains(distinct(L), Pred) :-
    closed_list(L),
    std_nonvar(Pred),
    Pred = V not_free_in T,
    !,
    member_eq(T, L),
    member_eq(V, L).


%% private prolog_code_contains_nfi(V1, T1, V2, T2).
%%
%%  Succeeds iff V1 not_free_in T1 implies V2 not_free_in T2.
%%
%%  mode prolog_code_contains_nfi(@obvar, @term, @obvar, @term).
%%
prolog_code_contains_nfi(V1, T1, V2, T2) :-
    V1 == V2,
    T1 == T2,
    !.
prolog_code_contains_nfi(V1, T1, V2, T2) :-
    V1 == T2,
    T1 == V2,
    !.
prolog_code_contains_nfi(V1, T1, V2, T2) :-
    list(T1),
    T1 = [T|TS],
    !,
    (
        prolog_code_contains_nfi(V1, T, V2, T2)
    ;
        prolog_code_contains_nfi(V1, TS, V2, T2)
    ),
    !.

    
%% private prolog_code_union(In1, In2, Out).
%%
%%  A very simplistic union whose main function is to eliminate
%%  unnecessary `true's.
%%
prolog_code_union(C1, C2, Out) :- C1 == true, !, Out = C2.
prolog_code_union(C1, C2, Out) :- C2 == true, !, Out = C1.
prolog_code_union(C1, C2, Out) :- Out = (C1,C2).


%% conjunction_to_list(Conjunction, CondList).
%% list_to_conjunction(CondList, Conjunction).
%%
%%  Eg.  Conjunction = 'ergo.and'(A, 'ergo.and'(B,C))
%%    and CondList = [A,B,C].
%%
%%  There is a unique mapping from condition lists to
%%  conjunctions, but the reverse mapping is not unique.
%%  For example, Conjunction='ergo.true' can mean CondList=[]
%%  or CondList=['ergo.true'].  Similarly, a 'ergo.and'/2 term in
%%  Conjunction can be split into several members of CondList, or one.
%%
%%  Thus, the implementation of conjunction_to_list is non-deterministic,
%%  but tries to make 'ergo.true' mean CondList=[] (at the outermost
%%  level only) and split as many conjunctions as possible.
%%  
%%  mode conjunction_to_list(@thy(term), ?list(thy(term))).
%%  mode list_to_conjunction(@list(thy(term)), ?thy(term)).
%%
conjunction_to_list(Term, []) :-
    Term == ERGO_TRUE.
conjunction_to_list(Term, CL) :-
    conjunction_to_list2(Term, CL).

conjunction_to_list2(Term, [C|CL]) :-
    compound(Term),
    functor(Term, Func, 2),
    Func == ERGO_AND,
    arg(1, Term, C),
    arg(2, Term, Term2),
    conjunction_to_list2(Term2,CL).
conjunction_to_list2(Term, [Term]).

list_to_conjunction([], ERGO_TRUE).
list_to_conjunction([C], C).
list_to_conjunction([C|CL], ERGO_AND(C,CS)) :-
    \+ CL = [],
    list_to_conjunction(CL,CS).


%% left_assoc_to_list(Term, Fun, List).
%% list_to_left_assoc(List, Term, Fun).
%% right_assoc_to_list(Term, Fun, List).
%% list_to_right_assoc(List, Term, Fun).
%%
%%  E.g. right_assoc_to_list(and(and(A,B),and(C,D)), and) --> [and(A,B),C,D].
%%  All are deterministic; the term-to-list versions produce maximal lists.
%%  The base cases are singleton lists; empty list is not handled.
%%
%%  mode left_assoc_to_list(@thy(term), @thy(term), ?list(thy(term))).
%%  mode list_to_left_assoc(@list(thy(term)), @thy(term), ?thy(term)).
%%  mode right_assoc_to_list(@thy(term), @thy(term), ?list(thy(term))).
%%  mode list_to_right_assoc(@list(thy(term)), @thy(term), ?thy(term)).
%% 
left_assoc_to_list(Term, Fun, List) :-
    left_assoc_to_list(Term, Fun, [], List).

left_assoc_to_list(Term, Fun, List0, List) :-
    compound(Term),
    Term = AFun(LT, RT),
    Fun == AFun,
    !,
    left_assoc_to_list(LT, Fun, [RT|List0], List).
left_assoc_to_list(Term, _, List0, [Term|List0]).

list_to_left_assoc([T], _, Term) :-
    !,
    Term = T.
list_to_left_assoc(List, Fun, Term) :-
    append(Front, [T], List),
    !,
    Term = Fun(LT, T),
    list_to_left_assoc(Front, Fun, LT).

right_assoc_to_list(Term, Fun, List) :-
    compound(Term),
    Term = AFun(LArg, RArg),
    Fun == AFun,
    !,
    List = [LArg|List1],
    right_assoc_to_list(RArg, Fun, List1).
right_assoc_to_list(Term, Fun, [Term]).

list_to_right_assoc([T], _, Term) :-
    !,
    Term = T.
list_to_right_assoc([T|Back], Fun, Term) :-
    Term = Fun(T, LT),
    list_to_right_assoc(Back, Fun, LT).

%% erase_all(Atom).
%%
%%  Erase all recorded facts with Atom as key.
%%
%%  mode erase_all(+atom).
%%
erase_all(A) :-
    recorded(A, _, Ref),
    erase(Ref),
    fail.
erase_all(_).
