%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: saveproof.ql,v 1.43 2000/11/15 06:29:47 itp Exp $


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file contains procedures for writing proofs to files.
%%
%% Note that there are several major problems at the moment!
%%
%% Firstly, when hypotheses are written out, they are topologically
%% sorted according to the (CDep) dependencies between them, so that a
%% rerun process can rerun them in the given order (otherwise it would 
%% have to do a similar sort itself).  However, this interacts badly 
%% with tactic wrappers, since it means that hypotheses often move out
%% of their wrappers.  For example, fwd often ends up as a wrapper
%% containing nothing (so when the proof is rerun, the fwd wrapper
%% has no effect and is discarded!).  One solution to this might be
%% to record the order in which conjectures are discharged, so that
%% rerun can do them in the same order.
%%
%% Another problem is that all the UDep stuff (which is used to
%% avoid saving hypothesis that are never used) is too complex and
%% creates large structures.  Better approaches might be (1) to go back
%% to saving all hypotheses (especially if we start sharing inherited
%% hypotheses rather than copying them),  (2) use a version of
%% get_theorem_dependencies to determine which hypotheses a given
%% subproof depends upon.
%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../lib/ergo.h"


save_proof parameter
  [set_save_proof, get_save_proof]
doc
  'The save_proof flag determines whether or not a proof will be',nl,
  'saved to a file.  The flag must be ''true'' (the default) or ''false''.'
extends
  parameters.


set_save_proof(Flag) :-
    check_or_error(
        member_var(Flag, [true, false]),
        ['illegal save proof flag: ',q(Flag)]),
    tactic_retractall_global(save_proof_record(_)),
    tactic_assert_global(save_proof_record(Flag)).

get_save_proof(Flag) :-
    ( query_global(save_proof_record(Fl)) ->
        Fl = Flag
    ;
        Flag = true
    ).




%% save_proof_record(+Theory, +TheoremName, +Type, +Term, +ProofTerm, +UDep)
%%
%%  Save a proof record to a file if the parameter save_proof is not set to
%%  false.
%%
%%  mode save_proof_record(@theory, @ground, @atom, @term, @proof, @udep_dms).
%%
save_proof_record(Theory, Name, Type, Term, Body, Proof, FU) :-
    ( get_save_proof( true ) ->
	save_proof_record_no_check(Theory, Name, Type, Term, Body, Proof, FU)
    ;
	true
    ).

save_proof_record_no_check(Theory, Name, Type, Term, Body, Proof, FU) :-
    %% assert dbg(print_proof) to turn on the following debug code.
    (dbg(print_proof) ->
        Proof = subprf(_,HL,FL,_),
        writelist(['Proof of: ',q(Name),'  type = ', q(Type), nl,
                   'Hyplist = ',nl,'    ']),
        pretty_print(HL,120,['   ']),
        writelist([nl,'Focuslist: ',nl,'   ']),
        pretty_print(FL,120,['   ']),
        writelist([nl, 'FocusUsage: ',nl,'   ']),
        pretty_print(FU,120,['   ']),
        writelist([nl])
    ;
        true
    ),
    ( get_save_proof(true) ->
        proof_file_name(Theory:Name, File),
        op_table(OldOpTable),
        obvar_prefix_table(OldPrefixTable),
        unwind_protect(
            save_proof_record2(Theory,Name,Type,Term,Body,Proof,FU,File),
            save_proof_record_cleanup(OldOpTable,OldPrefixTable))
    ;
        true
    ).

save_proof_record2(Theory,Name,Type,Term,Body,Proof,FU,File) :-
    ( file_type(Theory, dir) ->
        save_to(File)
    ;
        nl,
        write(' Creating proofs directory ...'),
        nl,
        make_directory(Theory),
        save_to(File)
    ), 
    !,
    op_table(proof),
    obvar_prefix_table(Theory),
    write_proof_record(Name, Term, Type, Proof, Body, FU).

save_proof_record_cleanup(OldOpTable,OldPrefixTable) :-
    saved,
    op_table(OldOpTable),
    obvar_prefix_table(OldPrefixTable).


%% private write_proof_record(...).
%%
write_proof_record(Name, Term, Type, subprf(_,HL,FL,_), Body, FU) :-
    write('proof_of('),
    writeq(Name), write(', '),
    writeq(Term), write(', '),
    writeq('2.0'), write(', '), 
    name(I, [10]),
    reverse(FL, FL2),
    reverse(FU, FU2),
    reverse(HL, HL2),
    write('['),
    write_proof_lines(FL2, FU2, HL2, I),
    write('],'),
    nl,
    writeq(Type),
    write(') :-'),
    nl,
    writeq(Body),
    write('.'),
    nl.



%% private write_proof(+Indent, +PStepList, +HypList, +UDep)
%%
%%   Write the proof of a window.
%%   This writes a list of the numbers of all inherited
%%   hypotheses that are used in the window, then
%%   writes the actual proof lines, which construct new
%%   hypotheses and transform the focus.
%%
write_proof(I, FL, HL, FU) :-
    reverse(FL, FL2),
    reverse(FU, FU2),
    reverse(HL, HL2),
    atom_append(I, ' ', J),
    ( get_inherited_hyps(HL2, FU, NH, HL3) ->
        true
    ;
        handle_error(program, [q(get_inherited_hyps(HL2,FU,NH,HL3))])
    ),
    write(NH),
    write(', ['), 
    write_proof_lines(FL2, FU2, HL3, J),
    write(']').


%% private write_proof_lines(+PStepList, +UDepList, +HypList, +Indent).
%%
%%  The first line in PStepList must be pstep(_,_,new),
%%  and this is ignored, together with the first member of UDepList.
%%
write_proof_lines([pstep(_,_,new)|Rest], [_|UDeps], HL, I) :-
    !,
    write_proof_lines2(Rest, UDeps, HL, I).
write_proof_lines(Lines, _, _, _) :-
    handle_error(program,['bad pstep list: ',q(Lines)]).

write_proof_lines2([], [], HL, I).
write_proof_lines2([pstep(N,F,A)|Rest], [U|RestU], HL, I) :-
    ( write_proof_line(A, U, HL, I) ->
        true
    ;
        handle_error(program,['pstep ',q(N),': ',thy(F),nl,
                              q(write_proof_line(A, U, HL, I))])
    ),        
    !,
    ( Rest = [] ->
        true
    ;
        write(',')
    ),
    write_proof_lines2(Rest, RestU, HL, I).


%% private write_proof_line(Just, UDep, HypList, Indent).
%%
write_proof_line(new, _, _, _).  %% ignore the initial focus.
write_proof_line(tr(R, D, G, A), U, HL, I) :-
    !,
    write_new_hyps_in(HL, U, I),
    write(I),
    writeq(tr(R, D, G, A)).
write_proof_line(op(P,subprf(_,H,F,_)), U, HL, I) :-
    !,
    write_new_hyps_in(HL, U, I),
    write(I),
    write('op('),
    write(P),
    write(', '), 
    write_proof(I, F, H, U),
    write(')').
write_proof_line(wr(Data, HNr, HNr2, subprf(_,_,F,_)), U, HL, I) :-
    select_hyps_in_range(HL, HNr, HNr2, InRng, HL2),
    write_new_hyps_in(HL2, U, I),
    write(I),
    write('wr('),
    writeq(Data),
    write(','),
    write(HNr),
    write(','),
    write(HNr2),
    write(', ['),
    atom_append(I, ' ', J),
    reverse(F, RF),
    reverse(U, RU),
    write_proof_lines(RF, RU, InRng, J),
    ( RF = [pstep(_,_,new)] ->
        %% RF has no lines to write.
        true
    ;
        write(',')
    ),
    write_new_hyps_in(InRng, U, I),
    write('nil'),
    write('])').


%% private get_inherited_hyps(+HypList0, +UDep, -INums, -HypList).
%%
%%  INums is the list of hypotheses (numbers only)
%%  which are mentioned in both HypList0 and UDep.
%%  HypList is all non-inherited hypotheses in HypList0.
%%
%%  Note: This assumes that all inherited hypotheses appear
%%        in HypList0 before any other hypotheses.  Thus,
%%        no inherited hypotheses will occur in the output HypList.
%%   
get_inherited_hyps([hrec(N, _, _, 'i', _, UNode)|HL], FU, Out, HL2) :-
    !,
    (udep_depends_on(FU, UNode) ->
        udep_delete(UNode),   %% Mark it as having been written.
        Out = [N|Out2]
    ;
        Out = Out2
    ),
    get_inherited_hyps(HL, FU, Out2, HL2).
get_inherited_hyps(HL, _, [], HL).


%% private write_new_hyps_in(+HypList, +UDep, +Indent).
%%
%%  Display all the hyps which appear in UDep (which is a list
%%  of udep DMS terms).
%%  Hyps are displayed prior to their use in other hyps
%%  (particularly conjectures).  This ensures that all the 
%%  hypotheses required to discharge a conjecture are
%%  present at the time the conjecture is first mentioned.
%%
%%  Each hypothesis that is written has its udep Node deleted to indicate
%%  that it has been written into this level of the proof.
%%  A comma is written after every hypothesis.
%%
write_new_hyps_in([], _, _) :-
    !.
write_new_hyps_in([hrec(N,T,_,cj(J),_,UNode)|HL], U, I) :-
    !,
    (udep_depends_on(U, UNode) ->
        udep_delete(UNode), % Mark it as having been written.
        udep_node_dependencies(UNode, U2),
        write_new_hyps_in(HL, U2, I),
        write_proof_conjecture(I, N, T, J, U2)
    ;
        true
    ),
    write_new_hyps_in(HL, U, I).
write_new_hyps_in([hrec(N,T,_,J,_,UNode)|HL], U, I) :-
    (udep_depends_on(U, UNode) ->
        udep_delete(UNode),  % Mark it as having been written.
        udep_node_dependencies(UNode, U2),
        write_proof_hyp(I,N,T,J,U2)
    ;
        true
    ),
    write_new_hyps_in(HL, U, I).


%% private write_proof_conjecture(...).
%%   
write_proof_conjecture(I, N, T, subprf(_,H,F,_), U) :-
    !,
    write(I),
    write('cj('),
    write(N),
    write(','),
    writeq(T),
    write(','), 
    write_proof(I, F, H, U),
    write('),').
write_proof_conjecture(I, N, T, J, _) :-
    PRE(write_proof_conjecture, J = tr(_,_,ERGO_TRUE,_)),
    write(I),
    writeq(cj(N,T,J)),
    write(',').


%% private write_proof_hyp(...).
%%
write_proof_hyp(I,N,_,oh(No,subprf(_,H,F,_)), U) :-
    !,
    write(I),
    write('oh('),
    write(N),
    write(','),
    write(No),
    write(','), 
    write_proof(I, F, H, U),
    write('),').
write_proof_hyp(I,N,T,J,_) :-
    write(I),
    hyp_proof_term(J,N,T,Hyp),
    writeq(Hyp),
    write(',').

%% private hyp_proof_term(+HypJust, +HypNum, +HypTerm, -SaveJust).
%%
hyp_proof_term(new,          N, T,  new(N, T)       ) :- !.
hyp_proof_term(tr(P),        N, T,  tr(P,N,T)       ) :- !.
hyp_proof_term(ht(No,R,D,A), N, T,  ht(No,R,D,T,A,N)) :- !.
hyp_proof_term(sl(No),       N, _,  sl(No,N)        ) :- !.
hyp_proof_term(sr(No),       N, _,  sr(No,N)        ) :- !.
hyp_proof_term(Just,         N, T,  _               ) :-
    handle_error(['unknown justification: ',q(Just),' in hyp ',
                  q(N),': ',thy(T)]).


%% private select_hyps_in_range(+HypList, +Low, +High, -InRange, -OutRange).
%%
select_hyps_in_range(HL,     A,   A,    [], HL) :-
    !.
select_hyps_in_range([],     _,   _,    [], []) :-
    !.
select_hyps_in_range([H|HL], Low, High, In, Out) :-
    PRE('select_hyps_in_range', Low < High),
    ( H = hrec(N,_,_,_,_,_), Low < N, N =< High ->
        In = [H|In2],
        Out = Out2
    ;
        In = In2,
        Out = [H|Out2]
    ),
    select_hyps_in_range(HL, Low, High, In2, Out2).


%% read_proof_record(SameVars,Thy,Name,Ver,ThmName,Ass,Term,Body,Proof).
%%
%%  Reads the proof record specified by Name, which must have
%%  the form Thy:RuleName, where Thy and RuleName are atoms.
%%
%%  If SameVars is true, then the names of variables in the proof record
%%  are remembered (this may cause them to be unified with named variables
%%  in the current state).  If it is false, then fresh names are used
%%  (this is the safest option if you are calling read_proof_record
%%  from within a proof).
%%  
%%  Ver is the version of the proof record format.
%%
%%  The remaining arguments all come from the original theorem
%%  that generated the proof record.
%%
%%  mode read_proof_record(@truefalse, @theory, @atom, ?atom, ?fullrulename,
%%                         ?list(term), ?term, ?code, ?proof).
%%
read_proof_record(Thy, Name, Ver, ThmName, [], Term, Body, Proof) :-
    read_proof_record(true, Thy, Name, Ver, ThmName, [], Term, Body, Proof).

read_proof_record(SameVars, Thy, Name, Ver, ThmName, [], Term, Body, Proof) :-
    check_truefalse(SameVars),
    proof_file_name(Thy:Name, File),
    get_load_path(PathList),
    member(P, PathList),
    full_file_name(P, File, _, PSFile),
    access(PSFile, 4, 0),  %% check that file is readable
    !,
    open(PSFile, read, Stream),
    unwind_protect(
        read_term(Stream, Clause,
                  [op_table(user),  %% Could change this to `proof' once
                                    %% all existing proofs are updated.
                   obvar_prefix_table(Thy),
                   remember_name(SameVars)]),
        close(Stream)),
    Clause = (proof_of(ThmName, Term, Ver, Proof, theorem) :- Body).
