%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: context.ql,v 1.36 2003/09/02 00:52:06 itp Exp $
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  This file contains all procedures for manipulating
%%  context terms in proof trees.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Implementation Notes:
%
%  Each node of a proof tree has zero or more 'context lists' associated
%  with it.  The terms in these context lists form the left hand side of
%  the sequent that the node represents.  Each kind of context has a
%  name, which is an atom.  Various procedures are provided to extend
%  and search context lists (most of them are in this file and have
%  names starting with context_...).
%
%  The data structures used to represent context lists are designed to:
%
%    * achieve a high degree of structure sharing (to save space, and
%      to save time by avoiding copying when creating new proof nodes);
%      Note that even different kinds of context lists (ie. with different
%      names) may share structure.
%
%    * allow efficient updating of any point in any context list;
%
%    * to allow changes to the context list at a given proof node to
%      be immediately visible at all nodes of its subtree.
%
%    * have changes undone automatically by Prolog failure
%      (this is important for writing efficient tactics)
%
%  In fact, the current implementation never copies context lists!
%  If a rule applies a filter function (like nfi(x)) or a transformation
%  function to a given context list, that context list is not immediately
%  filtered/transformed to obtain a new context list.  Instead, the
%  filter/transformation function is recorded at that point in the
%  context list, then whenever an access of the context list returns
%  a term T from before that point, it applies the filter/transformation
%  function to it.  In other words, the application of filter/transformation
%  functions is done lazily at lookup time, rather than at rule application
%  time.
%
%  Advantages of this approach are that it avoids copying of the context
%  list, it allows updates of the parent's context list to flow into
%  the `filtered/transformed' list automatically and it allow rule 
%  application to be very fast.
%
%  A disadvantage is that if a given term in the context list is
%  frequently accessed, then the filter/transformation function will
%  be applied to it each time, rather than just once.  If this turns
%  out to cause performance problems, it may be worthwhile caching
%  the filtered/transformed context list, or even switching to an
%  eager copy implementation.
%
%  The cspec notation within rules and the user-visible predicates
%  for manipulating context lists are carefully designed to allow
%  any of the following implementations:
%
%   1. Lazy application of filter/transformation functions, with full
%        sharing of context lists and no copying.
%   2. Lazy application, but with caching.
%   3. Eager application (always applies filter/transformation functions
%        immediately, and thus makes copies of context lists.
%
%  MU and PJR think that 1 will be best overall, but if benchmarking
%  shows that repeated application of filter/transformation functions
%  is a performance bottleneck for some inference rules, then adding
%  some form of caching may be a useful extension.  Note that, for some
%  logics, it is important that a cache system preserves the ordering
%  of terms within context lists.
%
%
%  Each term in a context list is stored as a (Id,Term) pair, where
%  Id is a 'name' (an integer actually) that is unique across the whole
%  proof and will not be reused upon backtracking.  These 'names' can
%  be used as identities for the context terms by user interfaces and
%  tactics.  The identities are allocated using the (non-backtrackable)
%  counter 'Context@TermId'.
%
%  All the context lists are stored in one huge forest of implicit
%  parameters, under the atom 'Context@Forest'.
%
%     IMPORTANT: this forest must only be accessed and
%                updated via procedures in this file.
%
%  The indexes into this forest are integers (called 'Cref's in the code),
%  allocated in ascending order from the global variable 'Context@Cref'.
%
%  Zero always points to the empty context list (forest_empty).
%  Other numbers point to various structures, whose semantics
%  is given by the following function M.
%  (We use the SVRC's Z-based specification language, Sum).
%
%    Cref == int;
%    Id   == int;       // the unique 'name' associated with each context term
%    [Term];            // theory terms
%    Obvar : power Term; // those theory terms that are object variables.
%    Test == power Term;  // user-defined tests over terms.
%
%    CNode ::= forest_empty          // never appears at root node of proof.
%            | forest_init(CName,seq(Id cross Term)) // only appears at root.
%            | forest_add(Id,Term,Cref)
%            | forest_delete(Id,Cref)
%            | forest_filter(Test,Cref)
%            | forest_transform(Func,Cref). (not supported yet)
%
%    F : Cref --> CNode;    // the forest (as at some given time)
%
%    M : CNode --> seq (Id cross Term);
%
%    M(forest_empty)          = <>;
%    M(forest_init(_,L))      = L;
%    M(forest_add(id,term,r)) = <(id,term)> ^ M(F(r));
%    M(forest_delete(id,r))   = squash {n |--> (i,t) : M(F(r)) | i /= id};
%    M(forest_filter(p,r))    = squash {n |--> (i,t) : M(F(r))
%                                       | p(t)};
%    M(forest_transform(p,f))    = squash {n |--> (i,t') | (i,t) : M(F(r))
%                                                          f(t,t')};
%
%  Intuitively, the context terms within a forest_init structure represent
%  context that is used by the proof, but is not generated within the
%  proof.  These context requirements are recorded in the inference rule
%  that results from the proof and when that rule is applied, the
%  required context must be present.  The second parameter of each
%  forest_init structure is actually an open list, but procedures like
%  context_search only *add* (Id,Term) pairs to it when called from
%  within a with_context_delaying/{1,2} procedure.
%
%  For each proof node N (where N is its identity) and each kind of context
%  CName, the implicit parameter CName(N) is bound to a Cref that points
%  into the context forest.  Thus, if ip_lookup(CName,N,Cref) is true
%  for some integer Cref, then the CName context list for node N
%  is M(F(Cref)).  If ip_lookup(CName,N,Cref) returns Cref as a variable,
%  then the CName context list for node N is the same as its parent's
%  CName context list.  This means that we can be lazy about doing
%  ip_set's for contexts that are unchanged.  This is particularly useful
%  when applying rules from a theory that has fewer context lists than
%  the current theory.
%  However, some benchmarking on large proofs would be useful to
%  determine whether it is faster to do ip_set's for unchanged context
%  lists (those that are known when the rule is defined) or not.
%  This probably depends upon the frequency of proof node construction
%  versus scanning down context lists.
%
%  Note that the crucial difference between context term identities
%  and cref's is that cref's are only used within this module, whereas
%  ctermid's are passed out so that tactics etc. can use them as identities.
%
%  Why have both?  Because some of the ways that we want to update
%  the forest (eg. list insertion) require moving a term from one
%  cref to another, whereas for external use we need a fixed 
%  relationship between terms and their identities (hence ctermids!).
%
%  

#include "../lib/ergo.h"

%% context_init.
%%
%%  This must be called before each proof, or perhaps before
%%  a cluster of related proofs within the same theory (if you want
%%  to view/edit parts of those proofs concurrently).
%%
%%  mode context_init.
%%
context_init :-
    get_proof_size(Size),
    ip_array_init('Context@Forest', Size),
    ip_set('Context@Forest', 0, forest_empty),
    ip_set('Context@Delay', false),
    global_state_set('Context@Cref', 0),
    global_state_set('Context@TermId', 0),
    findall(CName, is_context_name(CName), Contexts),
    %% like the rules in the database, we always keep
    %% the contexts in a canonical order (lexical).
    sort(Contexts, Contexts2),
    ip_set('Context@Contexts', Contexts2).


%% context_proof_init(PNode).
%%
%%  This must be called on the root node of each proof, before
%%  any proof steps are performed.
%%
%%  mode context_proof_init(@nodeid).
%%
context_proof_init(PNode) :-
    context_names(Contexts),
    context_proof_init2(Contexts, PNode).

context_proof_init2([], PNode).
context_proof_init2([CName|CS], PNode) :-
    global_state_increment('Context@Cref', Cref),
    %% extend the forest
    ip_set('Context@Forest', Cref, forest_init(CName,_)),
    get_proof_size(Size),
    ip_array_init(CName, Size),
    ip_set(CName, PNode, Cref),
    context_proof_init2(CS, PNode).


%% context_names(Names).
%%
%%  Names is the list of all context names defined in the current theory.
%%  This procedure should only be called while a proof is in progress.
%%
%%  mode context_names(?list(atom)).
%%
context_names(Names0) :-
    ip_lookup('Context@Contexts', Names),
    list(Names),
    Names = Names0.


cspec header
  'Rule context specification'
extends
  concepts
doc
  'A `cspec'' (Context Specification) is an expression that describes',nl,
  'the construction of a context list.  The syntax of cspecs is:',
  para,
  begin(verb),
  '  CSpec ::= empty',nl,
  '         |  use(CName,ProofNode)',nl,
  '         |  add(CIdVar,Term,CSpec)',nl,
  '         |  delete(CIdVar,Term,CSpec)',nl,
  '         |  filter(Func,CSpec).',nl,
  end(verb),nl,
  'Note that each CIdVar must be a metavariable.  This metavariable will',nl,
  'be instantiated to the identity of a context term when the cspec is',nl,
  'used to construct a context list.  This typically happens whenever an',nl,
  'inference rule is applied to a proof.',
  para,
  '`cdiff''s (Context Differences) are similar to cspecs, but are used to',nl,
  'describe the ',emph(difference),' between the context of two proof',nl,
  'nodes in a given proof, rather than as a specification for constructing',nl,
  'new context lists.  So `cdiff''s have exactly the same syntax as',nl,
  'cspecs, except that the CIdVar components are replaced by the actual',nl,
  'context term identities from a current proof.'.


library command
  context_diff(PNode, Ancestor, CName, CDiff)
extends 
  proof_tree_cmds
doc
  'CDiff is the difference between the CName context of proof nodes',nl,
  'PNode and Ancestor.  PNode must be within the subtree of Ancestor.',
  para,
  'CDiff is a `context difference'' expression, which is similar to a',nl,
  '`context specification'', but contains literal context term identifiers',nl,
  'from the proof, rather than variables for context term identifiers.',nl,
  'This is because it is more useful for user interfaces to see the',nl,
  'current reality of the proof.'
%%  To transform CDiff into a context',nl,
%%  'specification (by replacing each distinct context term identifier by a',nl,
%%  'unique metavariable) use ',
%%  xref('context_spec_abstract/3',global,cmd(context_spec_abstract/3)),'.'
mode
  context_diff(@proofnode, @proofnode, @cname, ?cdiff).

context_diff(PNode, Ancestor, CName, CDiff) :-
    context_lookup_cref(CName, PNode, Cref),
    integer(Cref),
    context_lookup_cref(CName, Ancestor, ACref),
    integer(ACref),
    !,
    context_diff2(Cref, Ancestor, [], CDiff).
context_diff(PNode, Ancestor, CName, _) :-
    handle_error(['bad args in ',q(context_diff(PNode,Ancestor,CName,...))]).


%% private context_diff2(Cref, Ancestor, DeletedCIds, CDiff).
%%
%%  mode context_diff2(@cref, @proofnode, @list(cmemid), -cdiff).
%%
context_diff2(Cref, Ancestor, Del, CDiff) :-
    %% see if Cref corresponds exactly to one of Ancestor's context lists.
    context_names(CNames),
    member(CN, CNames),
    context_lookup_cref(CN, Ancestor, Cref),
    !,
    context_diff_delete(Del, Cref, use(CN,Ancestor), CDiff).
context_diff2(Cref, Ancestor, Del, CDiff) :-
    %% it doesn't, so we scan down into Cref, building up CDiff.
    ip_lookup('Context@Forest', Cref, CStruct),
    context_diff3(CStruct, Ancestor, Del, CDiff).


%% private context_diff3(CStructure, Ancestor, Deleted, CDiff).
%%
%%  mode context_diff3(@cstruct, @proofnode, @list(cmemid), -cdiff).
%%
context_diff3(forest_empty, _, [], empty) :-
    !.
context_diff3(forest_empty, _, Del, _) :-
    handle_error(program, ['empty context should contain: ',q(Del)]).
context_diff3(forest_add(Id,T,Next), Ancestor, Del, CDiff) :-
    !,
    ( delete(Id, Del, Del2) ->
        %% it has been deleted further upstream, so we don't add it here
        CDiff = CDiff2
    ;
        Del2 = Del,
        CDiff = add(Id,T,CDiff2)
    ),
    context_diff2(Next, Ancestor, Del2, CDiff2).
context_diff3(forest_delete(Id,Next), Ancestor, Del, CDiff) :-
    !,
    context_diff2(Next, Ancestor, [Id|Del], CDiff).
context_diff3(forest_filter(Func,Next), Ancestor, Del, CDiff) :-
    !,
    CDiff = filter(Func, CDiff2),
    context_diff2(Next, Ancestor, Del, CDiff2).
%%context_diff3(forest_transform(Func,Next), Ancestor, Del, CDiff) :-
%%    !,
%%    CDiff = transform(Func, CDiff2),
%%    context_diff2(Next, Ancestor, Del, CDiff2).
context_diff3(CNode, Ancestor, _, _) :-
    %% if we reach here (eg. with forest_init), then Ancestor was incorrect
    %% (or the forest structures have been extended without changing this).
    handle_error(program, ['incorrect ancestor ',q(Ancestor),
                           ' to context_diff.  CNode=',q(CNode)]).

    
%% private context_diff_delete(Deleted, Cref, CDiffAncestor, CDiff).
%%
%%  CDiff extends CDiffAncestor by deleting all the context members
%%  whose identities are in Deleted.  All the deleted context members
%%  must be in the context list given by Cref.
%%
%%  mode context_diff_delete(@list(cmemid), @cref, ?cdiff, -cdiff).
%%
context_diff_delete([], _, CDiff, CDiff).
context_diff_delete([CId|Del], Cref, CDiffA, delete(CId,CTerm,CD)) :-
    context_search(Cref, CId, CTerm),
    !,
    context_diff_delete(Del, Cref, CDiffA, CD).
context_diff_delete([CId|_], Cref, _, _) :-
    handle_error(program, ['context ',q(Cref),' should contain: ',q(CId)]).


library command
  context_diff_to_spec(CDiff, VarMap, CSpec)
extends 
  proof_tree_cmds
doc
  'This procedure converts the `context difference'' expression CDiff into',nl,
  'a ',xref('`context specification''',global,cspec),
  ' expression CSpec, by replacing literal',nl,
  'context term identities (integers) with corresponding variables.',nl,
  'These variables will be instantiated back to new identities each time',nl,
  'the resulting CSpec is used to create a new context list.',
  para,
  'VarMap determines the replacements that will be performed to',nl,
  'obtain CSpec from CDiff.  VarMap determines the mapping from',nl,
  'context term identities to variables.  It is an open list containing',nl,
  'a cid(CId)=Var term for each CId in CDiff.  If necessary, these terms',nl,
  'will be added to VarMap by this procedure.  VarMap may also contain',nl,
  'proofnode(Id)=Var terms and these will be used to replace any literal',nl,
  'proof node identities that appear in use(CName,ProofNode) terms in CDiff.'
mode
  context_diff_to_spec(@cdiff, ?list(compound), ?cspec).

context_diff_to_spec(empty, _, empty).
context_diff_to_spec(use(CName,Node), VarMap, use(CName,Var)) :-
    search_insert_unify(proofnode(Node)=Var, VarMap).
context_diff_to_spec(add(Id,T,CDiff), VarMap, add(IdVar,T,CSpec)) :-
    search_insert_unify(cid(Id)=IdVar, VarMap),
    context_diff_to_spec(CDiff, VarMap, CSpec).
context_diff_to_spec(delete(Id,T,CDiff), VarMap, delete(IdVar,T,CSpec)) :-
    search_insert_unify(cid(Id)=IdVar, VarMap),
    context_diff_to_spec(CDiff, VarMap, CSpec).
context_diff_to_spec(filter(Func,CDiff), VarMap, filter(Func,CSpec)) :-
    context_diff_to_spec(CDiff, VarMap, CSpec).

%% TO DO: search_insert_unify should be called search_insert.

%% TO DO: index on second arg in search_insert_unify?

search_insert_unify(Val, List) :-
    var(List),
    !,
    List = [Val|_].
search_insert_unify(Val, [Val|_]) :-
    !.
search_insert_unify(Val, [_|List]) :-
    search_insert_unify(Val, List).


%% context_constraints(Root, VarMap, SearchList).
%%
%%  SearchList is a list of context_search/4 calls that will be executed
%%  during rule application to check that all the delayed/assumed
%%  context terms of proof node Root are present.  VarMap should be
%%  the same map that is passed to context_diff_to_spec.
%%
%%  mode context_constraints(@proofnode, ?list(compound), -list(compound)).
%%
context_constraints(Root, VarMap, SearchList) :-
    context_names(CNames),
    search_insert_unify(proofnode(Root)=RootVar, VarMap),
    map(context_constraints2(Root,RootVar,VarMap,SearchList), [CNames]),
    open_to_closed(SearchList).

context_constraints2(Root, RootVar, VarMap, SearchList, CName) :-
    context_lookup_cref(CName, Root, Cref),
    ip_lookup('Context@Forest', Cref, forest_init(CName, L)),
    %% insert the context_search/4's for CName into the open SearchList.
    context_constraints3(L, RootVar, VarMap, SearchList, CName).

context_constraints3(L, _, _, _, _) :-
    var(L),
    !.
context_constraints3([(Id,T)|L], RootVar, VarMap, SearchList, CName) :-
    search_insert_unify(cid(Id)=IdVar, VarMap),
    search_insert(context_search(CName, RootVar, IdVar, T), SearchList),
    context_constraints3(L, RootVar, VarMap, SearchList, CName).



library command
  with_context_delaying(Code),
  with_context_delaying(Patt, Code)
extends 
  proof_tree_cmds
doc
  'These procedures change the behaviour of most procedures that search',nl,
  'context lists so that they add the given context term if it is not',nl,
  'already present in the context.  This makes it easy to prove anything!',nl,
  'However, all such added context terms are recorded as constraints of',nl,
  'each inference rule generated from the proof.  This is typically useful',nl,
  'for proving inference rules that manipulate context lists.',
  para,
  'This changed behaviour applies only during the execution of Code.',nl,
  'It is allowable for Code to contain nested calls to these procedures.',
  para,
  'Patt provides more control over what context terms can be added.',nl,
  'Only those that are unifiable with Patt can be added.'
mode
  with_context_delaying(+code),
  with_context_delaying(@thy(term), +code).

with_context_delaying(Code) :-
    with_context_delaying(_, Code).

with_context_delaying(Patt, Code) :-
    ip_lookup('Context@Delay', Orig),
    ip_set('Context@Delay', patt(Patt)),
    call_predicate(Code),
    ip_set('Context@Delay', Orig).


library command
  context_assume(CName, PNode, CId, CTerm)
extends 
  proof_tree_cmds
doc
  'This procedure ensures that CId:::CTerm is available within the CName',nl,
  'context of PNode.  If it is not already present, and the CName',nl,
  'context of PNode is inherited from the root node of the proof, then',nl,
  'CId:::CTerm is added to the root node of the proof as an assumption.'
mode
  context_assume(@cname, @proofnode, ?ctermid, ?thy(term)).

context_assume(CName, PNode, CId, CTerm) :-
    %% This is equivalent to with_delaying_context(context_search(...)).
    ip_lookup('Context@Delay', Orig),
    ip_set('Context@Delay', patt(_)),
    context_search(CName, PNode, CId, CTerm),
    ip_set('Context@Delay', Orig).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The rest of this file defines various 'rule_constraint' commands
%% that can be called from the constraints section of inference rules
%% to manipulate/search context lists.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


rule_constraint command
  context_search(CName, PNode, CId, CTerm)
extends
  rule_constraint_cmds
doc
  'Succeeds when the CName context of the proof node PNode',nl,
  'contains CTerm (CId is an identifier for that occurence of CTerm).'
mode
  context_search(@cname, @nodeid, ?ctermid, ?thy(term)).

context_search(CName, PNode, CId, CTerm) :-
    context_lookup_cref(CName, PNode, Cref),
    ip_lookup('Context@Forest', Cref, CNode),
    context_search2(CNode, CId, CTerm).


%% context_search(Cref, CId, CTerm).
%%
%%  A lower level equivalent to context_search/4, for use from
%%  within the core prover (and inference rules) in cases when
%%  we want to start from a Cref, rather than from a proof node.
%%
%%  mode context_search(@cref, ?ctermid, ?thy(term)).
%%  
context_search(Cref, CId, CTerm) :-
    ip_lookup('Context@Forest', Cref, CNode),
    context_search2(CNode, CId, CTerm).


%% private.
%%
context_search2(L, Id, T) :-
    var(L),
    !,
    ip_lookup('Context@Delay', Permission),
    compound(Permission),
    Permission = patt(Patt),
    Patt ?= T,
    %% allocate CId
    global_state_increment('Context@TermId', Id),
    L = [(Id,T)|_].
context_search2([(Id,T)|_], Id, T1) :- rule_unify(T, T1).
context_search2([_|L], Id, T) :-
    context_search2(L, Id, T).
context_search2(forest_add(Id,T,_Next), Id, T1) :- rule_unify(T, T1).
context_search2(forest_add(_,_,Next), CId, CTerm) :- 
    ip_lookup('Context@Forest', Next, CNode),
    context_search2(CNode, CId, CTerm).
context_search2(forest_delete(Id,Next), CId, CTerm) :-
    ip_lookup('Context@Forest', Next, CNode),
    context_search2(CNode, CId, CTerm),
    CId =\= Id.
context_search2(forest_filter(nfi(Var),Next), CId, CTerm) :-
    ip_lookup('Context@Forest', Next, CNode),
    context_search2(CNode, CId, CTerm),
    Var not_free_in CTerm.
context_search2(forest_filter(Func,Next), CId, CTerm) :-
    \+ functor(Func, nfi, _),
    ip_lookup('Context@Forest', Next, CNode),
    context_search2(CNode, CId, CTerm),
    Func(CTerm).
%%context_search2(forest_transform(Func,Next), CId, TCTerm) :-
%%    ip_lookup('Context@Forest', Next, CNode),
%%    context_search2(CNode, CId, CTerm),
%%    call_predicate(Func, CTerm, TCTerm). 
context_search2(forest_init(_,L), Id, T) :-
    context_search2(L, Id, T).



rule_constraint command
  context_index(CName, PNode, Index, CId)
extends
  rule_constraint_cmds
doc
  'Succeeds when the CName context of the proof node PNode',nl,
  'contains a term with identifier CId at index Index.'
mode
  context_index(@cname, @nodeid, ?number, ?ctermid).

context_index(CName, PNode, Index, CId) :-
    context_lookup_cref(CName, PNode, Cref),
    ip_lookup('Context@Forest', Cref, CNode),
    context_index_aux(CNode, CId, Index, 1, _, []).

context_index_aux(XXX, _, _, _, _, _) :- var(XXX),!,fail.
context_index_aux([], _, _, Index, Index, _).
context_index_aux([(Id, _)|_], Id, Index, Index, _, Deleted) :-
    \+ member(Id, Deleted),
    !.
context_index_aux([(Id, _)|Rest], FId, FIndex, Index, NewIndex, Deleted) :-
    member(Id, Deleted),
    !,
    context_index_aux(Rest, FId, FIndex, Index, NewIndex, Deleted).
context_index_aux([(_Id, _)|Rest], FId, FIndex, Index, NewIndex, Deleted) :-
    Index1 is Index+1, 
    context_index_aux(Rest, FId, FIndex, Index1, NewIndex, Deleted).
context_index_aux(forest_add(Id,_T,_Next), Id, Index, Index, _NewIndex, Deleted) :-
    \+ member(Id, Deleted),
    !.
context_index_aux(forest_add(Id,_T,Next), FId, FIndex, Index, NewIndex, Deleted) :-
    member(Id, Deleted),
    ip_lookup('Context@Forest', Next, CNode),
    !,
    context_index_aux(CNode, FId, FIndex, Index, NewIndex, Deleted).
context_index_aux(forest_add(_Id,_T,Next), FId, FIndex, Index, NewIndex, Deleted) :-
    ip_lookup('Context@Forest', Next, CNode),
    Index1 is Index+1, 
    context_index_aux(CNode, FId, FIndex, Index1, NewIndex, Deleted).
context_index_aux(forest_delete(Id,Next), FId, FIndex, Index, NewIndex, Deleted) :-
    ip_lookup('Context@Forest', Next, CNode),
    context_index_aux(CNode, FId, FIndex, Index, NewIndex, [Id|Deleted]).
context_index_aux(forest_filter(nfi(_Var),Next), FId, FIndex, Index, NewI, Deleted) :-
    ip_lookup('Context@Forest', Next, CNode),
    context_index_aux(CNode, FId, FIndex, Index, NewI, Deleted).
context_index_aux(forest_init(_,L), FId, FIndex, Index, NewIndex, Deleted) :-
    context_index_aux(L, FId, FIndex, Index, NewIndex, Deleted).

% Retrieve a context term given the context name, the proof node ID
% and the context ID. This predicate does not apply context constraints.
% It is simply used to look up a given context term.
%
% mode context_retrieve(@cname, @nodeid, @ctermid, ?thy(term)).

context_retrieve(CName, PNode, CId, CTerm) :-
    nonvar(CName), nonvar(PNode), nonvar(CId),
    context_lookup_cref(CName, PNode, Cref),
    ip_lookup('Context@Forest', Cref, CNode),
    context_retrieve2(CNode, CId, CTerm).

% private

context_retrieve2(L, _, _) :- var(L),!,fail.
context_retrieve2([(Id,T)|_], Id, T) :- !.
context_retrieve2([_|L], Id, T) :-
    context_retrieve2(L, Id, T).
context_retrieve2(forest_add(Id,T,Next), Id, T).
context_retrieve2(forest_add(_,_,Next), CId, CTerm) :-
    ip_lookup('Context@Forest', Next, CNode),
    context_retrieve2(CNode, CId, CTerm).
context_retrieve2(forest_delete(Id,Next), CId, CTerm) :-
    ip_lookup('Context@Forest', Next, CNode),
    context_retrieve2(CNode, CId, CTerm),
    CId =\= Id.
context_retrieve2(forest_filter(nfi(_),Next), CId, CTerm) :-
    ip_lookup('Context@Forest', Next, CNode),
    context_retrieve2(CNode, CId, CTerm).
context_retrieve2(forest_filter(Func,Next), CId, CTerm) :-
    \+ functor(Func, nfi, _),
    ip_lookup('Context@Forest', Next, CNode),
    context_retrieve2(CNode, CId, CTerm).
%%context_retrieve2(forest_transform(Func,Next), CId, TCTerm) :-
%%    ip_lookup('Context@Forest', Next, CNode),
%%    context_search2(CNode, CId, CTerm),
%%    call_predicate(Func, CTerm, TCTerm). 
context_retrieve2(forest_init(_,L), Id, T) :-
    context_retrieve2(L, Id, T).



%% context_lookup_cref(CName, PNode, Cref).
%%
%%  This is equivalent to ip_lookup(CName, PNode, Cref), except
%%  that if the resulting Cref is a variable (because the CName
%%  context of PNode has not been set yet) then it searches up the
%%  parent links of PNode and returns the first nonvar Cref.
%%
%%  mode context_lookup_cref(@cname, @proofnode, -cref).
%%
context_lookup_cref(CName, PNode, Cref) :-
    ip_lookup(CName, PNode, Cref0),
    integer(Cref0),
    !,
    Cref = Cref0.
context_lookup_cref(CName, PNode, Cref) :-
%% TO DO: Use QP's inline to achieve the same effect as the following in
%% context_lookup_cref?
    %% WARNING: here is an inline version of proof_node_parent(PNode,Parent).
    %% This breaks the modularity of proof.ql, but it is important for
    %% this lookup to be efficient, because searching context is common.
    ip_lookup('Proof@Node', PNode, proofnode(Parent, _, _, _, _, _, _)),
    integer(Parent),
%% TO DO: Can we avoid the following overkill in context_lookup_cref?
    %% We set this node's CName pointer so that future searches
    %% will not have to scan up the parent pointers.
    %% Note that we do this for *all* the parents that haven't had
    %% their CName context set yet, which is probably overkill because
    %% context searches are typically only initiated from open proof nodes.
    %%
    ip_set(CName, PNode, Cref),
    %%
    context_lookup_cref(CName, Parent, Cref).



rule_constraint command
  context_nfi(ObVar,PNode)
extends
  rule_constraint_cmds
doc
  'Adds an nfi(ObVar) filter to ',emph(all),' the context lists of',nl,
  'proof node PNode.  This effectively removes all terms T in those',nl,
  'context lists that do not satisfy ObVar not_free_in T.'
mode
  context_nfi(@obvar, @nodeid).

context_nfi(ObVar, PNode) :-
    PRE(context_nfi, obvar(ObVar)),
    context_names(Contexts),
    map(context_nfi(ObVar, PNode), [Contexts]).

context_nfi(ObVar, PNode, CName) :-
    PRE(context_nfi, obvar(ObVar)),
    global_state_increment('Context@Cref', Cref),
    %% extend the forest
    context_lookup_cref(CName, PNode, Cref0),
    %% Remove any previous nfi(ObVar)
    %%% Include either the following two lines or the line after that.
    % context_nfi_remove_nfi(ObVar, Cref1, Cref0),
    % ip_set('Context@Forest', Cref, forest_filter(nfi(ObVar),Cref1)),
    %%%
    ip_set('Context@Forest', Cref, forest_filter(nfi(ObVar),Cref0)),
    %%%
    ip_set(CName, PNode, Cref).

%% TO DO: Fix context_nfi_remove_nfi to remove double filters
%% and reinsert call in context_nfi.

%% context_nfi_remove_nfi removes any previous filters nfi(ObVar).

% context_nfi_remove_nfi(ObVar, forest_empty, forest_empty).
% context_nfi_remove_nfi(ObVar, Removed, forest_filter(nfi(X),Present)) :-
%     ObVar == X,
%     !,
%     context_nfi_remove_nfi(ObVar, Removed, Present).
% context_nfi_remove_nfi(ObVar, forest_init(I,L), forest_init(I,L)) :-
%     !.
% context_nfi_remove_nfi(ObVar, forest_add(Id,Removed), forest_add(Id,Present)) :-
%     !,
%     context_nfi_remove_nfi(ObVar, Removed, Present).
% context_nfi_remove_nfi(ObVar, forest_delete(Id,Removed), forest_delete(Id,Present)) :-
%     !,
%     context_nfi_remove_nfi(ObVar, Removed, Present).
% context_nfi_remove_nfi(ObVar, Present, Present).

