%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: ruleconstraints.ql,v 1.8 2006/02/24 02:32:24 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Standard constraint procedures for inference rules.
%%
%% WARNING:  It is important for soundness that all of these are
%%           monotonic with respect to Qu-Prolog instantiation.
%%           That is, if a constraint succeeds on a term T,
%%           it should also succeed on all instantiations of T.
%%           So, nonvar(_) is monotonic, but var(_) is not!
%%
%%           If you want non-monotonic constraints, then you
%%           must leave delayed checks (see delay_until below),
%%           so that the success is conditional on those checks
%%           succeeding later.  Ergo will collect up all such
%%           delayed checks into each theorem that you prove.
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rule_constraint_cmds header
  'Constraints on rules'
extends
  concepts
doc
  'The following commands can be used to add constraints on the',nl,
  'application of inference rules (and theorems).',nl.


rule_constraint command
  defined(D,E)
extends
  rule_constraint_cmds
doc
  'True if D is defined to be equal to E in an ancestor theory.'
mode
  defined(?thy(term),?thy(term)).

defined(D,E) :-
    nonvar(D),
    ( compound(D) -> functor(D, F, _), nonvar(F)
    ; quant(D) -> quantify(D,F,Bnd,_), nonvar(F), closed_list(Bnd)
    ; atomic(D), F = D
    ),
    !,
    %% we can go straight to the correct theory.
    internal_name(F, Thy, _),
    query_theory(rule(_, defn, _, 'ergo.='(DD,E), NFIs), Thy),
    rule_unify(DD,D),
    call_predicate(NFIs).
defined(D,E) :-
    nonvar(E),
    ( compound(E) -> functor(E, F, _), nonvar(F)
    ; quant(E) -> quantify(E,Q,Bnd,_), nonvar(Q), closed_list(Bnd)
    ; true
    ),
    !,
    query_ancestors(rule(_, defn, _, 'ergo.='(D,EE), NFIs)),
    rule_unify(EE,E),
    call_predicate(NFIs).
defined(D,E) :-
    collect_vars((D,E), V),
    filter(var, V, [X|_]),
    delay(X, defined(D,E)).

rule_constraint command
  integer(I)
extends
  rule_constraint_cmds
doc
  'True if I is a literal integer constant.'
mode
  integer(?integer).


rule_constraint command
  atom(A)
extends
  rule_constraint_cmds
doc
  'True if A is an atom, like ''abc''.'
mode
  atom(?atom).


rule_constraint command
  not_free_in(Obvar,Term)
extends
  rule_constraint_cmds
doc
  'True if Obvar does not appear free in Term.',nl,
  'If this is not decidable, delayed constraints will be added to',nl,
  'ensure that Obvar can never appear free in Term.'
mode
  not_free_in(?thy(obvar), ?thy(term)).

rule_constraint command
  distinct(Obvars)
extends
  rule_constraint_cmds
doc
  'True if Obvars is a list of distinct object variables.',nl,
  'If this is not decidable, delayed constraints will be added to',nl,
  'ensure that the object variables are distinct.'
mode
  distinct(?list(thy(obvar))).


rule_constraint command
  is_not_free_in(Obvar,Term)
extends
  rule_constraint_cmds
doc
  'True if Obvar obviously does not appear free in Term.',nl,
  'Unlike the `not_free_in'' constraint, this constraint never adds',nl,
  'delayed constraints, so if the not_free_in check is not decidable,',nl,
  'this predicate will fail.'
mode
  is_not_free_in(?thy(obvar), ?thy(term)).


rule_constraint command
  '='(Term1,Term2)
extends
  rule_constraint_cmds
doc
  'True if Term1 can be unified with Term2.',nl,
  'If the two terms are unifiable, but there is no unique unifier,',nl,
  'then delayed unification constraints will be added to the proof.',nl,
  'For example, [A/x]B = 1 has two possible unifiers:  B=1, or B=x,A=1,',nl,
  'so this unification would be delayed.'
mode
  '='(?thy(term), ?thy(term)).


rule_constraint command
  delay_until(Condition, Constraints)
extends
  rule_constraint_cmds
doc
  'True iff Constraints is true.',nl,
  'Constraints can be any other sequence rule constraints.',nl,
  'However, checking of those constraints is delayed until',nl,
  'Condition is true.  This is typically used to allow proof steps',nl,
  'to succeed when the proof terms contain variables, but leave',nl,
  'constraints on the future instantiations of those variables.',
  para,
  'Here is a hypothetical example:',
  para,
  '     delay_until(ground([I,J]), (integer(I),integer(J),I<J))',
  para,
  'checks that I<J, but delays that check until both I and J are',nl,
  'fully instantiated to ground terms (containing no variables).',nl,
  'This might be useful in an arithmetic theory, if you wanted to prove',nl,
  'a theorem like  param_const(intconst,I) < param_const(intconst,J),',nl,
  'for all literal values of I and J.',
  para,
  'The possible values of Condition are:',nl,
  begin(list),
  item('nonvar(X)'), '  delay until X becomes a non-variable.',nl,
  item('ground(X)'), '  delay until X becomes ground (containing no variables)',nl,
  item('bound(X)'),  '  delay until the variable (or object variable) X is',nl,
  'bound to any term (including another variable or object variable).',nl,
  '(Not very useful in rule constraints)',nl,
  item('identical_or_apart(X, Y)'), '  delay until the terms X and Y',nl,
  'become either identical or non-unifiable.',nl,
  item('or(A,B)'), '  delay until either of the delay conditions',nl,
  'A or B is satisfied.',nl,
  item('and(A,B)'),'  delay until both of the delay conditions A and B',nl,
  'are satisfied.',
  end(list)
mode
  delay_until(+nonvar, +rule_constraint).

