%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: jprop.thy,v 1.46 2006/02/01 23:36:26 itp Exp $
%%
%% Johansson's Minimal Logic
%% =========================
%%
%% This is an Ergo theory of Johansson's minimal logic [1].
%% Minimal logic is an intuitionistic logic in which implication is
%% more narrowly interpreted than in Heyting's intuitionistic logic.
%%
%% References:
%%   [1] I. Johansson. Der Minimalkalk\"ul, ein Reduzierter Intuitionistischer
%%       Formalismus.  Compositio Mathematica, 4, pp 119-136, 1936.
%%       
%%       This introduces minimal logic and gives the differences between
%%       it and various formulations of intuitionistic logic.
%%
%%       A.  From Heyting's axioms for intuitionistic logic [3].
%%
%%           Drop axiom 4.1:  not(P) => (P => Q).
%%
%%       B.  From Gentzen's NJ (intuitionistic natural deduction) logic [4].
%%
%%           Drop the inference rule:
%%
%%               false
%%             ---------
%%                 P
%%
%%           and treat not(P) as an abbreviation for P => false.
%%
%%       C.  From Gentzen's LJ (intuitionistic sequent calculus) logic [4].
%%
%%   [2] A. Heyting. Die Formalen Regeln der Intuitionistischen Logik.
%%       Sitzungsberichte der Deutschen Akademie der Wissenschaften zu Berlin,
%%       Klasse fur Mathematik, Physik und Technik. pp 42-56, 1930. 
%%
%%   [3] A. Heyting. Intuitionism: An Introduction.  Studies in Logic and the
%%       Foundations of Mathematics. North-Holland. 1971.
%%
%%       Pages 105-106 describe the differences between Heyting's
%%       intuitionistic logic and Johansson's minimal logic.
%%       The difference is that minimal logic drops axiom 10:
%%
%%          not(A) => (A => B)
%%
%%   [4] Gerhard Gentzen.  "Investigations into Logical Deduction."
%%       Chapter 3 in "The Collected Papers of Gerhard Gentzen",
%%       Edited by M. E. Szabo, North-Holland, 1969.  Pages 68-131.
%%       First published in German in 1934.
%%
%%   [5] Peter J. Robinson and John Staples.  "Formalizing the
%%       Hierarchical Structure of Practical Mathematical Reasoning".
%%       Journal of Logic and Computation, Vol 3, Num 1, 1993.
%%       Pages 47-61.
%%
%% The approach taken here is to use Gentzen's NJ (intuitionistic
%% natural deduction) logic [4], adapted as described above ([1] B).
%% We make not/1 a function, rather than an Ergo abbreviation, so that
%% we can prove derived inference rules for it, and thus present a
%% more abstract interface to descendent theories.
%%
%% Rationale: This approach gives an easy starting point for proving 
%% the various inference rules that we want in order to support 
%% window inference [5], and also makes it easy to prove Heyting's ten 
%% axioms for minimal logic [3].  In contrast, starting with just the
%% axioms, plus a few very basic inference rules (eg. modus ponens),
%% would make it harder to prove the other inference rules.
%% 
%% We make not/1 a function, rather than an Ergo abbreviation, so that
%% we can prove derived inference rules for it, and thus present a
%% more abstract interface to descendent theories.

add_theory(jprop).
include_theory(stdops, share).

declare false.
declare _ and _.
declare _ or _.
declare _ => _.       % This is called "implies" or "imp" within rule names.


%% Encoding Natural Deduction Assumptions into Ergo
%% ================================================
%%
%% The assumptions that arise in natural deduction proofs are treated
%% in a sequent-calculus style within Ergo, by defining a context called
%% 'hyp' (for 'hypotheses') that contains the set of assumptions that are
%% valid at each point in the proof.  As usual [4], the semantics of a
%% proof node:
%%
%%     hyp=[A1,A2,...An] ---> Concl
%%
%% is that the conjunction of the 'hyp' context terms, A1..An, must
%% imply the conclusion of the sequent, Concl.
%%
%% In addition to the usual natural deduction inference rules, which
%% only add hypotheses, we provide the 'assump' rule which defines
%% how hypotheses can be used to discharge predicates and the 'hyp_delete'
%% rule which allows unwanted hypotheses to be removed.  Note that
%% we do not need all the structural inference rules of Gentzen's
%% sequent calculus, because Ergo allows us to view context lists
%% as unordered sets of terms.

add_context_name(hyp).

axiom assump(CId:ctermid) ===
    ID ::: C
  provided
    context_search(hyp, ID, CId, C).

axiom hyp_delete(CId:ctermid,CTerm:thy(term)) === 
    ID1 ::: hyp---[CId ::: CTerm] ---> C
    ----------------------------------------
    ID0 ::: C
  provided
    context_search(hyp, ID0, CId, CTerm).

%%
%% inference rules for and/2.
%% =========================
%%

axiom and_intro ===
    A,
    B
    -------
    A and B.

axiom and__r(A:thy(term)) ===
    A and B
    -------
    B.

axiom and__l(B:thy(term)) ===
    A and B
    -------
    A.

%%
%% rules for or/2.
%% ===============
%%

axiom l__or ===
    A
    ----------
    A or B.

axiom r__or ===
    B
    ----------
    A or B.

axiom or_elim(A:thy(term),B:thy(term)) ===
    A or B,
    hyp+++[A] ---> C,
    hyp+++[B] ---> C
    -----------------
    C.


%%
%% rules for =>/2.
%% ===============
%%

axiom imp_intro ===
    hyp+++[A] ---> B
    ----------------
    A => B.

axiom imp_elim(A:thy(term)) ===
    A => B,
    A
    -------
    B.


%%
%% Support for unfolding and folding definitions.
%% =============================================
%%


axiom expansion_elim(P:thy(term),x:obvar,D:thy(term)) ===
    [E/x]P
    ----------
    [D/x]P
  provided defined(D,E).


axiom expansion_intro(P:thy(term),x:obvar,D:thy(term)) ===
    [D/x]P
    ----------
    [E/x]P
  provided defined(D,E).

close_theory.



define not P === (P => false).

define true === not false.


%% We read <=> as "is equivalent to" or "iff".  Rule names use "iff".
%%
define A <=> B === (A => B) and (B => A).

%% NB. We need '<=' ("follows from") to support window inference.
%%     However, defining <= gives two notations for the same concept,
%%     which can quickly lead to an explosion in the number of rules
%%     and thus reduce proof efficency.  To avoid this, we follow the 
%%     principle that the *only* rules that should use <= are the
%%     reflexivity, transitivity and opening rules that are needed to
%%     support window inference.  As far as possible, all other rules
%%     should be expressed using =>.
%%
%%     "<=" is called "follows" within rule names.
%%
define A <= B === B => A.

%% It is convenient to prove a few context manipulation rules first.
%%


theorem iff_expansion(P:thy(term),x:obvar,D:thy(term)) ===
   [D/x]P <=> [E/x]P
 provided defined(D,E).

theorem expansion_iff(P:thy(term),x:obvar,D:thy(term)) ===
   [E/x]P <=> [D/x]P
 provided defined(D,E).

theorem and_elim_c(A:thy(term), B:thy(term)) ===
    hyp---[CId ::: A and B]+++[A,B] ---> C
    ---------------------------------------
    Node :::                             C
  provided
    context_search(hyp, Node, CId, A and B).

theorem or_elim_c(B:thy(term), A:thy(term)) ===
    hyp---[CId ::: B or A]+++[B] ---> C,
    hyp---[CId ::: B or A]+++[A] ---> C
    ---------------------------
    ConcId ::: C
  provided
    context_search(hyp, ConcId, CId, B or A).

theorem imp_assump(CId:ctermid) ===
    hyp---[CId ::: B => A]+++[A] ---> C,
    hyp+++[]                     ---> B
    ------------------------------------
    Node :::                          C
  provided
    context_search(hyp, Node, CId, B => A).

theorem modus_ponens_c(CId:ctermid) ===
        B
    ----------
    Node ::: A
  provided
    context_search(hyp, Node, CId, B => A).


theorem imp_elim_c(B:thy(term), A:thy(term)) ===
    hyp---[CId1 ::: B => A]+++[A] ---> C
    ------------------------------------
    Node :::                          C
  provided
    context_search(hyp, Node, CId1, B => A),
    context_search(hyp, Node, CId2, B).

%% The theorems/axioms of Johansson's minimal logic are provable.
theorem and_intro_r ===
    A => A and A.

theorem sym_and2 ===
    A and B => B and A.

theorem and_elim_r ===
    A and B => B.

theorem and_elim_l ===
    A and B => A.

theorem and_idempotent ===
    (A and A) <=> A.

theorem and_extract ===
        A and (B and C) <=> B and (A and C).                                    

theorem and_imp ===
    (A => B) => (A and C => B and C).

theorem trans_imp ===  %% was called transimp
    (A => B) and (B => C) => (A => C).

theorem imp_imp ===
    A => (B => A).

theorem modus_ponens ===
    A and (A => B) => B.

theorem modus_ponens2 ===
    (A => B) and A => B.

theorem or_intro_r ===
    A => (A or B).

theorem sym_or2 ===
    A or B => B or A.

theorem imp_or_intro ===
    (A => B) and (C => B) <=> (A or C => B).

theorem not_intro ===
    (A => B) and (A => not B) => not A.


%% Reflexivity and Transitivity of the three preorder relations,
%% plus the strength relationships between them.
%%
theorem imp_reflex ===
    A => A.

theorem fol_reflex ===
    A <= A.

theorem iff_reflex ===
    A <=> A.

theorem imp__imp__imp(C:thy(term)) ===
    B => C,
    C => A
    ----------
    B => A.

theorem fol__fol__fol(C:thy(term)) ===
    B <= C,
    C <= A
    ----------
    B <= A.

theorem iff__iff__iff(F:thy(term)) ===
    B <=> F,
    F <=> A
    ----------
    B <=> A.

theorem iff__imp ===
    A <=> B
    ----------
    A => B.

theorem iff__fol ===
    A <=> B
    ----------
    A <= B.

%% Conversion from follows to implies

theorem imp__fol ===
    A => B
    ------
    B <= A.


%% Window inference opening rules.
%% Our naming convention is: open_<function>_<position>_<relation>

% openrule(A and B, [1], 
%         [B],
%         id,
%         [=> ---> =>,
%          =  ---> <=>],
%         [A and B ---> A,
%          true    ---> true]).

theorem open_and_1_imp ===
    hyp+++[A] ---> C => B
    ------------------------------
    (C and A) => (B and A).

theorem open_and_1_fol ===
    hyp+++[A] ---> C <= B
    ------------------------------
    (C and A) <= (B and A).

theorem open_and_1_iff ===
    hyp+++[B] ---> A <=> A2
    ------------------------------
    (A and B) <=> (A2 and B).


% openrule(A and B, [2],
%         [A],
%         id,
%         [=> ---> =>,
%          =  ---> <=>],
%         [A and B ---> B,
%          true    ---> true]).

theorem open_and_2_imp ===
    hyp+++[B] ---> C => A
    ------------------------------
    (B and C) => (B and A).

theorem open_and_2_fol ===
    hyp+++[B] ---> C <= A
    ------------------------------
    (B and C) <= (B and A).

theorem open_and_2_iff ===
    hyp+++[A] ---> B <=> B2
    ------------------------------
    (A and B) <=> (A and B2).


% openrule(A or B, [1],
%         [], 
%         id,
%         [=> ---> =>,
%          =  ---> <=>],
%         [A or B ---> A,
%          true   ---> true]).

theorem open_or_1_imp ===
    C => B
    ------------------------------
    (C or A) => (B or A).

theorem open_or_1_fol ===
    C <= B
    ------------------------------
    (C or A) <= (B or A).

theorem open_or_1_iff ===
    A <=> A2
    ------------------------------
    (A or B) <=> (A2 or B).


% openrule(A or B, [2],
%         [],
%         id,
%         [=> ---> =>,
%          =  ---> <=>],
%         [A or B ---> B,
%          true   ---> true]).

theorem open_or_2_imp ===
    C => A
    ------------------------------
    (B or C) => (B or A).

theorem open_or_2_fol ===
    C <= A
    ------------------------------
    (B or C) <= (B or A).

theorem open_or_2_iff ===
    B <=> B2
    ------------------------------
    (A or B) <=> (A or B2).


% openrule(A => B,  [1], 
% 	 [], 
% 	 id, 
% 	 [=> ---> rev(=>),
%           =  ---> <=>],  
% 	 [A => B ---> A, 
% 	  not A  ---> A]).

theorem open_imp_1_imp ===
    C <= B
    ------------------------------
    (C => A) => (B => A).

theorem open_imp_1_fol ===
    C => B
    ------------------------------
    (C => A) <= (B => A).

theorem open_imp_1_iff ===
    C <=> B
    ------------------------------
    (C => A) <=> (B => A).


% openrule(A => B,  [2],
%         [A], 
%         id,
%         [=> ---> =>,
%          = ---> <=>],
%         [A => B ---> B,
%          true   ---> true]).

theorem open_imp_2_imp ===
    hyp+++[B] ---> C => A
    ------------------------------
    (B => C) => (B => A).

theorem open_imp_2_fol ===
    hyp+++[B] ---> C <= A
    ------------------------------
    (B => C) <= (B => A).

theorem open_imp_2_iff ===
    hyp+++[B] ---> C <=> A
    ------------------------------
    (B => C) <=> (B => A).


% openrule(not A,   [1],
% 	 [],
% 	 id,
% 	 [=> ---> rev(=>),
%           =  ---> <=>],  
% 	 [not A ---> A]).

theorem open_not_1_imp ===
    B <= A
    --------------------
    not B => not A.

theorem open_not_1_fol ===
    B => A
    --------------------
    not B <= not A.

theorem open_not_1_iff ===
    B <=> A
    --------------------
    not B <=> not A.


% openrule(A <=> B, [1], 
%         [],
%         id,
%         [<=> ---> <=>,
%          =   ---> <=>],
%         [A <=> B ---> A, 
%          true    ---> B]).

theorem open_iff_1_iff ===
    C <=> B
    -----------------------
    (C <=> A) <=> (B <=> A).

theorem open_iff_1_imp ===
    C <=> B
    -----------------------
    (C <=> A) => (B <=> A).

theorem open_iff_1_fol ===
    C <=> B
    -----------------------
    (C <=> A) <= (B <=> A).


% openrule(A <=> B, [2], 
%         [],
%         id,
%         [<=> ---> <=>,
%          =   ---> <=>],
%         [A <=> B ---> B,
%          true    ---> A]).

theorem open_iff_2_iff ===
    B <=> B2
    ------------------------------
    (A <=> B) <=> (A <=> B2).

theorem open_iff_2_imp ===
    C <=> B
    -----------------------
    (A <=> C) => (A <=> B).

theorem open_iff_2_fol ===
    C <=> B
    -----------------------
    (A <=> C) <= (A <=> B).




%% Symmetry (Commutativity) rules
%% ==============================

theorem and_iff_and ===
    A and B <=> B and A.

theorem and__and ===
    A and B
    -------
    B and A.

theorem or_iff_or ===
    A or B <=> B or A.

theorem or__or ===
    A or B
    ------
    B or A.

theorem iff_IFF_iff ===
    (A <=> B) <=> (B <=> A).

theorem iff__iff ===
    A <=> B
    -------
    B <=> A.

theorem eqv_comm ===
  (A <=> B) => (B <=> A). 

%% Distribution rules
%% ==================

theorem deMorgan_or ===
    not (B or A) => not B and not A.

theorem deMorgan_and ===
    not A or not B => not (A and B).

theorem and_or ===
    (A and B) or C <=> (A or C) and (B or C).

theorem or_and ===
    A or (B and C) <=> (A or B) and (A or C).

theorem or_and_2 ===
    (A or B) and C <=> (A and C) or (B and C).

theorem and_or_2 ===
    A and (B or C) <=> (A and B) or (A and C).

theorem and_and ===
    (A and B) and C <=> (A and C) and (B and C).

theorem and_and1 ===
    A and (B and C) <=> (A and B) and (A and C).

theorem or_or ===
    (A or B) or C <=> (A or C) or (B or C).

theorem or_or1 ===
    B or (C or A) <=> (B or C) or (B or A).

theorem imp_and ===
    (A => (B and C)) <=> (A => B) and (A => C).


%% Associativity rules
%% ===================

theorem and_assoc ===
    A and (B and C) <=> (A and B) and C.

theorem and_and__and_and ===
    ( A and B ) and C
    -----------------
    A and ( B and C ).

theorem and_AND__AND_and ===
    A and ( B and C )
    -----------------
    ( A and B ) and C.

theorem or_assoc ===
    C or (B or A) <=> (C or B) or A.

theorem or_or__or_or ===
    ( A or B ) or C
    ---------------
    A or ( B or C ).

theorem or_OR__OR_or ===
    A or ( B or C )
    ---------------
    ( A or B ) or C.


%% Simplification rules
%% ====================

theorem or_elim_iff ===  %% was called or_elim
    A or A <=> A.



%% Theorems about true
%% ===================

theorem true ===
    true.

theorem not_not_true ===
  not not true.  

theorem true_and ===
    true and A <=> A.

theorem and_true ===
    A and true <=> A.

theorem or_true ===
    A or true <=> true.

theorem true_or ===
    true or A <=> true.

theorem true_imp ===
    (true => A) <=> A.

theorem imp_true ===
    A => true.

theorem true_iff ===
    (true <=> A) <=> A.

theorem iff_true ===
    (A <=> true) <=> A.

%% Other rules
%% ===========

theorem weak_contradiction ===
    A and not A => not B.

theorem not_intro1 ===
    (A => B and not B) => not A.

theorem or_intro_l ===
    B => A or B.

theorem case_or ===
    (A or B) and (A => C) and (B => C) 
    => C.

theorem mp ===
    A => ((A => B) <=> B).

theorem antecedents ===
    (A => (B => C)) <=> (A and B => C).

theorem strengthen ===
    (A => B) => ((A => C) <=> (B and A => C)).

theorem strengthen2 ===
    (A => B) => ((A => C) <=> (B => (A => C))).

theorem imp_not_not ===
    A => not not A.

theorem not_iff_not_not_not ===
    not A <=> not not not A.

theorem non_contradiction ===
    not (A and not A).

theorem not_not_or ===
    not not (A or not A).

theorem imp_not_and ===
    (A => B) => not (A and not B).

theorem imp_iff_and ===
    (A => not B) <=> not (A and B).

theorem imp_iff_imp ===
    (not not A => B) => (not B => not A).

theorem not_imp_not_and_l ===
    not A => not (A and B).

theorem not_imp_not_and_r ===
    not B => not (A and B).

theorem iff_true__l ===
    A <=> true
    ----------
    A.

theorem true_iff__r ===
    true <=> A
    ----------
    A.

theorem imp__imp__iff ===
    A => B,
    B => A
    -------
    A <=> B.

theorem iff_imp_l ===
    (A <=> B) => (A => B).

theorem iff_imp_r ===
    (A <=> B) => (B => A).

theorem iff_mp_l ===
    A and (A <=> B) => B.

theorem iff_mp_r ===
    B and (A <=> B) => A.

theorem r__iff__l ===
    B,
    A <=> B
    -------
    A.

theorem l__iff__r ===
    A,
    A <=> B
    -------
    B.

theorem l__imp__r ===
    A,
    A => B
    ------
    B.

theorem r__fol__l ===
    B,
    A <= B
    ------
    A.

theorem imp_imp_and ===
    A => (B => A and B).

theorem imp_imp_and1 ===
    A => (B => B and A).

theorem equiv_conj ===
  (A and B <=> A and C) <=> (A => (B <=> C)).
 
theorem equiv_conj_l ===
  (A and B <=> A) <=> (A => B).
 
theorem equiv_conj_r ===
  (A <=> A and B) <=> (A => B).                                               

%% Extra Unsorted Theorems
%% =======================

theorem from__id__not( CIdA:ctermid, CIDnotA:ctermid, A:thy(term) ) ===
    ConcId ::: not B
  provided
    context_search( hyp, ConcId, CIdA, A ),
    context_search( hyp, ConcId, CIdnotA, not A ).

theorem from__id__imp( CIdAss:ctermid, CIdImp:ctermid, B:thy(term) ) ===
    ConcId ::: A
  provided
    context_search( hyp, ConcId, CIdAss, B ),
    context_search( hyp, ConcId, CIdImp, B => A ).

theorem id__not_not ===
    A
    ---------
    not not A.

theorem l__and_true ===
    A
    ----------
    A and true.

theorem r__true_and ===
    A
    ----------
    true and A.

%% TO DO: Think up good names for the following.
%theorem l__or_true ===
%    A or true.
%
%theorem l__true_or ===
%    true or A.

theorem r__true_imp ===
    A
    ---------
    true => A.

theorem l__iff_true ===
    A
    ----------
    A <=> true.

theorem r__true_iff ===
    A
    ----------
    true <=> A.

theorem hyp_imp_intro( B:thy(term), A:thy(term) ) ===
    hyp---[CBimpA ::: B=>A]+++[A] ---> C
    -----------------------------
    ConcId ::: C
  provided
    context_search(hyp, ConcId, CBimpA, B=>A),
    context_search(hyp, ConcId, CB, B).

theorem hyp_iff_intro( B:thy(term), A:thy(term) ) ===
    hyp---[CBiffA ::: B<=>A]+++[B=>A, A=>B] ---> C
    -----------------------------
    ConcId ::: C
  provided
    context_search(hyp, ConcId, CBiffA, B<=>A).

theorem hyp_not_intro( A:thy(term) ) ===
    hyp---[CnotA ::: not A]+++[A=>false] ---> B
    -----------------------------
    ConcId ::: B
  provided
    context_search(hyp, ConcId, CnotA, not A).

theorem fold_not ===
    A => false
    ----------
    not A.

theorem hyp_add(B:thy(term)) ===
    hyp+++[B] ---> A,
    B
    ------------
    A.

theorem hyp_add2(CId:ctermid, B:thy(term)) ===
    B,
    hyp+++[CId ::: B] ---> A
    ------------------------
    A. 

theorem hyp_replace(CA:ctermid,A:thy(term),B:thy(term)) ===
    hyp---[CA ::: A]+++[B] ---> C,
    B
    ------------
    ConcId ::: C
  provided
    context_search(hyp, ConcId, CA, A).


%% Theorems to support Window Inference tactics


theorem fol__imp ===
        B <= A
        ------
        A => B.

theorem fol_true__r ===
        A <= true
        ---------
            A.

theorem hyp_imp(New:ctermid,Old:ctermid) ===
      hyp---[Old:::H] ---> (H => K),
      hyp+++[New:::K] ---> G
      ----------------------
      C_0:::G
    provided context_search(hyp, C_0, Old, H).

theorem and_hyp_l(H:ctermid, NewH:ctermid) ===
      hyp+++[NewH:::A] ---> G
      -----------------------
              C_0:::G
   provided context_search(hyp, C_0, H, A and B).

theorem and_hyp_r(H:ctermid, NewH:ctermid) ===
      hyp+++[NewH:::B] ---> G
      -----------------------
              C_0:::G
   provided context_search(hyp, C_0, H, A and B).

theorem l__fol_true ===
      A
   -------
   A <= true.

theorem l__imp_true ===
     A
  ------
  A => true.

theorem iff_rt ===
        A,
        A <=> B
        ----------
        true <=> B.

theorem imp_rt ===
        A,
        A => B
        ----------
        true => B.

theorem fol_rt ===
        A,
        A <= B
        ----------
        true <= B.

theorem iff_rt_ante ===
        H => A,
        H,
        A <=> B
        ----------
        true <=> B.

theorem imp_rt_ante ===
        H => A,
        H,
        A => B
        ----------
        true => B.

theorem fol_rt_ante ===
        H => A,
        H,
        A <= B
        ----------
        true <= B.

theorem fol_intro ===
        hyp+++[B] ---> A
        --------
        A <= B.


%%
%% Support for Window Inference - loads
%% jprop specific information.
%%

require_global_tactic_file(wi_support).
require_tactic_file(wi_support,jprop).




